""" Test class for Job Cleaning Agent
"""

# imports
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pytest
from mock import MagicMock

# DIRAC Components
from DIRAC.WorkloadManagementSystem.Agent.JobCleaningAgent import JobCleaningAgent
from DIRAC import gLogger

gLogger.setLevel("DEBUG")

# Mock Objects
mockReply = MagicMock()
mockAM = MagicMock()
mockNone = MagicMock()
mockNone.return_value = None
mockJMC = MagicMock()


@pytest.fixture
def jca(mocker):
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobCleaningAgent.AgentModule.__init__")
    mocker.patch(
        "DIRAC.WorkloadManagementSystem.Agent.JobCleaningAgent.AgentModule._AgentModule__moduleProperties",
        side_effect=lambda x, y=None: y,
        create=True,
    )
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobCleaningAgent.AgentModule.am_getOption", return_value=mockAM)
    mocker.patch(
        "DIRAC.WorkloadManagementSystem.Agent.JobCleaningAgent.JobDB.getDistinctJobAttributes", side_effect=mockReply
    )
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobCleaningAgent.JobDB.selectJobs", side_effect=mockReply)
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobCleaningAgent.JobDB.__init__", side_effect=mockNone)
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobCleaningAgent.ReqClient", return_value=mockNone)
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobCleaningAgent.JobMonitoringClient", return_value=mockJMC)

    jca = JobCleaningAgent()
    jca.log = gLogger
    jca.log.setLevel("DEBUG")
    jca._AgentModule__configDefaults = mockAM
    jca.initialize()

    return jca


@pytest.mark.parametrize(
    "mockReplyInput, expected",
    [
        ({"OK": True, "Value": ""}, {"OK": True, "Value": []}),
        ({"OK": False, "Message": ""}, {"OK": False, "Message": ""}),
    ],
)
def test__getAllowedJobTypes(jca, mockReplyInput, expected):
    """Testing JobCleaningAgent()._getAllowedJobTypes()"""

    mockReply.return_value = mockReplyInput
    result = jca._getAllowedJobTypes()
    assert result == expected


@pytest.mark.parametrize(
    "mockReplyInput, expected",
    [
        ({"OK": True, "Value": ""}, {"OK": True, "Value": None}),
        ({"OK": False, "Message": ""}, {"OK": False, "Message": ""}),
    ],
)
def test_removeJobsByStatus(jca, mockReplyInput, expected):
    """Testing JobCleaningAgent().removeDeletedJobs()"""

    mockReply.return_value = mockReplyInput
    result = jca.removeDeletedJobs()
    assert result == expected


@pytest.mark.parametrize(
    "conditions, mockReplyInput, expected",
    [
        ({"JobType": "", "Status": "Deleted"}, {"OK": True, "Value": ""}, {"OK": True, "Value": None}),
        ({"JobType": "", "Status": "Deleted"}, {"OK": False, "Message": ""}, {"OK": False, "Message": ""}),
        ({"JobType": [], "Status": "Deleted"}, {"OK": True, "Value": ""}, {"OK": True, "Value": None}),
        (
            {"JobType": ["some", "status"], "Status": ["Deleted", "Cancelled"]},
            {"OK": True, "Value": ""},
            {"OK": True, "Value": None},
        ),
    ],
)
def test_deleteJobsByStatus(jca, conditions, mockReplyInput, expected):
    """Testing JobCleaningAgent().deleteJobsByStatus()"""

    mockReply.return_value = mockReplyInput
    result = jca.deleteJobsByStatus(conditions)
    assert result == expected


@pytest.mark.parametrize(
    "inputs, params, expected",
    [
        ([], {"OK": True, "Value": {}}, {"OK": True, "Value": {"Failed": {}, "Successful": {}}}),
        (["a", "b"], {"OK": True, "Value": {}}, {"OK": True, "Value": {"Failed": {}, "Successful": {}}}),
        (
            [],
            {"OK": True, "Value": {1: {"OutputSandboxLFN": "/some/lfn/1.txt"}}},
            {"OK": True, "Value": {"Failed": {}, "Successful": {1: "/some/lfn/1.txt"}}},
        ),
        (
            [],
            {
                "OK": True,
                "Value": {1: {"OutputSandboxLFN": "/some/lfn/1.txt"}, 2: {"OutputSandboxLFN": "/some/other/lfn/2.txt"}},
            },
            {"OK": True, "Value": {"Failed": {}, "Successful": {1: "/some/lfn/1.txt", 2: "/some/other/lfn/2.txt"}}},
        ),
        (
            ["a", "b"],
            {"OK": True, "Value": {1: {"OutputSandboxLFN": "/some/lfn/1.txt"}}},
            {"OK": True, "Value": {"Failed": {}, "Successful": {1: "/some/lfn/1.txt"}}},
        ),
        (["a", "b"], {"OK": False}, {"OK": False}),
    ],
)
def test_deleteJobOversizedSandbox(mocker, inputs, params, expected):
    """Testing JobCleaningAgent().deleteJobOversizedSandbox()"""

    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobCleaningAgent.AgentModule.__init__")
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobCleaningAgent.AgentModule.am_getOption", return_value=mockAM)
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobCleaningAgent.JobDB", return_value=mockNone)
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobCleaningAgent.ReqClient", return_value=mockNone)
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.JobCleaningAgent.JobMonitoringClient", return_value=mockJMC)

    jobCleaningAgent = JobCleaningAgent()
    jobCleaningAgent.log = gLogger
    jobCleaningAgent.log.setLevel("DEBUG")
    jobCleaningAgent._AgentModule__configDefaults = mockAM
    jobCleaningAgent.initialize()

    mockJMC.getJobParameters.return_value = params

    result = jobCleaningAgent.deleteJobOversizedSandbox(inputs)

    assert result == expected



""" Basic unit tests for the Job API
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from os.path import dirname, join

import pytest
from six import StringIO

from DIRAC.Interfaces.API.Job import Job
from DIRAC.Core.Utilities.ClassAd.ClassAdLight import ClassAd


def test_basicJob():
    job = Job()

    job.setOwner("ownerName")
    job.setOwnerGroup("ownerGroup")
    job.setName("jobName")
    job.setJobGroup("jobGroup")
    job.setExecutable("someExe")
    job.setType("jobType")
    job.setDestination("ANY")

    xml = job._toXML()

    with open(join(dirname(__file__), "testWF.xml")) as fd:
        expected = fd.read()

    assert xml == expected

    with open(join(dirname(__file__), "testWFSIO.jdl")) as fd:
        expected = fd.read()

    jdlSIO = job._toJDL(jobDescriptionObject=StringIO(job._toXML()))
    assert jdlSIO == expected


def test_SimpleParametricJob():

    job = Job()
    job.setExecutable("myExec")
    job.setLogLevel("DEBUG")
    parList = [1, 2, 3]
    job.setParameterSequence("JOB_ID", parList, addToWorkflow=True)
    inputDataList = [
        ["/lhcb/data/data1", "/lhcb/data/data2"],
        ["/lhcb/data/data3", "/lhcb/data/data4"],
        ["/lhcb/data/data5", "/lhcb/data/data6"],
    ]
    job.setParameterSequence("InputData", inputDataList, addToWorkflow=True)

    jdl = job._toJDL()

    with open(join(dirname(__file__), "testWF.jdl")) as fd:
        expected = fd.read()

    assert jdl == expected

    clad = ClassAd("[" + jdl + "]")

    arguments = clad.getAttributeString("Arguments")
    job_id = clad.getAttributeString("JOB_ID")
    inputData = clad.getAttributeString("InputData")

    assert job_id == "%(JOB_ID)s"
    assert inputData == "%(InputData)s"
    assert "jobDescription.xml" in arguments
    assert "-o LogLevel=DEBUG" in arguments
    assert "-p JOB_ID=%(JOB_ID)s" in arguments
    assert "-p InputData=%(InputData)s" in arguments


@pytest.mark.parametrize(
    "proc, minProc, maxProc, expectedProc, expectedMinProc, expectedMaxProc",
    [
        (4, None, None, 4, None, 4),
        (4, 2, None, 4, None, 4),
        (4, 2, 8, 4, None, 4),
        (4, 8, 6, 8, None, 8),  # non-sense
        (None, 2, 8, None, 2, 8),
        (None, 1, None, None, 1, None),
        (None, None, 8, None, 1, 8),
        (None, 8, 8, 8, None, 8),
        (None, 12, 8, 8, None, 8),  # non-sense
    ],
)
def test_MPJob(proc, minProc, maxProc, expectedProc, expectedMinProc, expectedMaxProc):

    job = Job()
    job.setExecutable("myExec")
    job.setLogLevel("DEBUG")
    job.setNumberOfProcessors(proc, minProc, maxProc)
    jdl = job._toJDL()
    clad = ClassAd("[" + jdl + "]")
    processors = clad.getAttributeInt("NumberOfProcessors")
    minProcessors = clad.getAttributeInt("MinNumberOfProcessors")
    maxProcessors = clad.getAttributeInt("MaxNumberOfProcessors")
    assert processors == expectedProc
    assert minProcessors == expectedMinProc
    assert maxProcessors == expectedMaxProc



""" unit tests for Transformation Clients
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=missing-docstring,blacklisted-name,invalid-name

import unittest
import importlib
from mock import MagicMock

from DIRAC import S_OK, gLogger
from DIRAC.TransformationSystem.Client.TaskManagerPlugin import TaskManagerPlugin


class opsHelperFakeUser(object):
    def getValue(self, foo="", bar=""):
        if foo == "JobTypeMapping/AutoAddedSites":
            return ["CERN", "IN2P3"]
        # This is for 'JobTypeMapping/<jobType>/AutoAddedSites
        elif foo.endswith("AutoAddedSites"):
            return bar
        # and this is for 'JobTypeMapping/<jobType>/Exclude
        return ["PAK", "Ferrara", "Bologna", "Paris", "Hospital"]

    def getOptionsDict(self, foo=""):
        # The only call to this method is for /JobTypeMapping/<type>/Allow
        return {"OK": True, "Value": {"Paris": "IN2P3"}}


class opsHelperFakeUser2(object):
    def getValue(self, foo="", bar=""):
        if foo == "JobTypeMapping/AutoAddedSites":
            return ""
        # This is for 'JobTypeMapping/<jobType>/AutoAddedSites
        elif foo.endswith("AutoAddedSites"):
            return bar
        # and this is for 'JobTypeMapping/<jobType>/Exclude
        return ["PAK", "Ferrara", "Bologna", "Paris", "CERN", "IN2P3", "Hospital"]

    def getOptionsDict(self, foo=""):
        return {"OK": True, "Value": {"Paris": "IN2P3", "CERN": "CERN", "IN2P3": "IN2P3"}}


class opsHelperFakeDataReco(object):
    def getValue(self, foo="", bar=""):
        if foo == "JobTypeMapping/AutoAddedSites":
            return ["CERN", "IN2P3"]
        # This is for 'JobTypeMapping/<jobType>/AutoAddedSites
        elif foo.endswith("AutoAddedSites"):
            return bar
        # and this is for 'JobTypeMapping/<jobType>/Exclude
        return ["PAK", "Ferrara", "CERN", "IN2P3", "Hospital"]

    def getOptionsDict(self, foo=""):
        return {"OK": True, "Value": {"Ferrara": "CERN", "IN2P3": "IN2P3, CERN"}}


class opsHelperFakeHospital(object):
    def getValue(self, foo="", bar=""):
        if foo == "JobTypeMapping/AutoAddedSites":
            return ["CERN", "IN2P3"]
        # This is for 'JobTypeMapping/<jobType>/AutoAddedSites
        elif foo.endswith("AutoAddedSites"):
            return []
        # and this is for 'JobTypeMapping/<jobType>/Exclude
        return ["ALL"]

    def getOptionsDict(self, foo=""):
        return {"OK": True, "Value": {"Hospital": "CERN, IN2P3"}}


class opsHelperFakeMerge(object):
    def getValue(self, foo="", bar=""):
        if foo == "JobTypeMapping/AutoAddedSites":
            return ["CERN", "IN2P3", "Hospital"]
        # This is for 'JobTypeMapping/<jobType>/AutoAddedSites
        elif foo.endswith("AutoAddedSites"):
            return bar
        # and this is for 'JobTypeMapping/<jobType>/Exclude
        return ["ALL"]

    def getOptionsDict(self, foo=""):
        return {"OK": False, "Message": "JobTypeMapping/MCSimulation/Allow in Operations does not exist"}


class opsHelperFakeMerge2(object):
    def getValue(self, foo="", bar=""):
        if foo == "JobTypeMapping/AutoAddedSites":
            return ""
        # This is for 'JobTypeMapping/<jobType>/AutoAddedSites
        elif foo.endswith("AutoAddedSites"):
            return bar
        # and this is for 'JobTypeMapping/<jobType>/Exclude
        return ["ALL"]

    def getOptionsDict(self, foo=""):
        return {"OK": True, "Value": {"CERN": "CERN", "IN2P3": "IN2P3"}}


class opsHelperFakeMC(object):
    def getValue(self, foo="", bar=""):
        if foo == "JobTypeMapping/AutoAddedSites":
            return ["CERN", "IN2P3"]
        # This is for 'JobTypeMapping/<jobType>/AutoAddedSites
        elif foo.endswith("AutoAddedSites"):
            return bar
        # and this is for 'JobTypeMapping/<jobType>/Exclude
        return [""]

    def getOptionsDict(self, foo=""):
        return {"OK": False, "Message": "JobTypeMapping/MCSimulation/Allow in Operations does not exist"}


def getSitesFake():
    return S_OK(["Ferrara", "Bologna", "Paris", "CSCS", "PAK", "CERN", "IN2P3", "Hospital"])


def getSitesForSE(ses):
    if ses == ["CERN-DST"] or ses == "CERN-DST":
        return S_OK(["CERN"])
    elif ses == ["IN2P3-DST"] or ses == "IN2P3-DST":
        return S_OK(["IN2P3"])
    elif ses == ["CSCS-DST"] or ses == "CSCS-DST":
        return S_OK(["CSCS"])
    elif ses == ["CERN-DST", "CSCS-DST"] or ses == "CERN-DST,CSCS-DST":
        return S_OK(["CERN", "CSCS"])


#############################################################################


class ClientsTestCase(unittest.TestCase):
    """Base class for the clients test cases"""

    def setUp(self):

        gLogger.setLevel("DEBUG")

        self.mockTransClient = MagicMock()
        self.mockTransClient.setTaskStatusAndWmsID.return_value = {"OK": True}

        self.WMSClientMock = MagicMock()
        self.jobMonitoringClient = MagicMock()
        self.mockReqClient = MagicMock()

        self.jobMock = MagicMock()
        self.jobMock2 = MagicMock()
        mockWF = MagicMock()
        mockPar = MagicMock()
        mockWF.findParameter.return_value = mockPar
        mockPar.getValue.return_value = "MySite"

        self.jobMock2.workflow = mockWF
        self.jobMock2.setDestination.return_value = {"OK": True}
        self.jobMock.workflow.return_value = ""
        self.jobMock.return_value = self.jobMock2

        self.maxDiff = None

    def tearDown(self):
        pass


#############################################################################


class TaskManagerPluginSuccess(ClientsTestCase):
    def test__BySE(self):

        ourPG = importlib.import_module("DIRAC.TransformationSystem.Client.TaskManagerPlugin")
        ourPG.getSitesForSE = getSitesForSE
        p_o = TaskManagerPlugin("BySE", operationsHelper=MagicMock())

        p_o.params = {"Site": "", "TargetSE": ""}
        res = p_o.run()
        self.assertEqual(res, set([]))

        p_o.params = {"Site": "ANY", "TargetSE": ""}
        res = p_o.run()
        self.assertEqual(res, set([]))

        p_o.params = {"TargetSE": "Unknown"}
        res = p_o.run()
        self.assertEqual(res, set([]))

        p_o.params = {"Site": "Site1;Site2", "TargetSE": ""}
        res = p_o.run()
        self.assertEqual(res, set([]))

        p_o.params = {"TargetSE": "CERN-DST"}
        res = p_o.run()
        self.assertEqual(res, set(["CERN"]))

        p_o.params = {"TargetSE": "IN2P3-DST"}
        res = p_o.run()
        self.assertEqual(res, set(["IN2P3"]))

        p_o.params = {"TargetSE": "CERN-DST,CSCS-DST"}
        res = p_o.run()
        self.assertEqual(res, set(["CERN", "CSCS"]))

        p_o.params = {"TargetSE": ["CERN-DST", "CSCS-DST"]}
        res = p_o.run()
        self.assertEqual(res, set(["CERN", "CSCS"]))

    def test__ByJobType(self):

        ourPG = importlib.import_module("DIRAC.TransformationSystem.Client.TaskManagerPlugin")
        ourPG.getSites = getSitesFake
        ourPG.getSitesForSE = getSitesForSE

        # "User" case - 1
        p_o = TaskManagerPlugin("ByJobType", operationsHelper=opsHelperFakeUser())

        p_o.params = {"Site": "", "TargetSE": "CERN-DST", "JobType": "User"}
        res = p_o.run()
        self.assertEqual(res, set(["CERN", "CSCS", "IN2P3"]))

        p_o.params = {"Site": "", "TargetSE": "IN2P3-DST", "JobType": "User"}
        res = p_o.run()
        self.assertEqual(res, set(["IN2P3", "Paris", "CSCS", "CERN"]))

        p_o.params = {"Site": "", "TargetSE": ["CERN-DST", "CSCS-DST"], "JobType": "User"}
        res = p_o.run()
        self.assertEqual(res, set(["CERN", "CSCS", "IN2P3"]))

        # "User" case - 2
        p_o = TaskManagerPlugin("ByJobType", operationsHelper=opsHelperFakeUser2())

        p_o.params = {"Site": "", "TargetSE": "CERN-DST", "JobType": "User"}
        res = p_o.run()
        self.assertEqual(res, set(["CERN", "CSCS"]))

        p_o.params = {"Site": "", "TargetSE": "IN2P3-DST", "JobType": "User"}
        res = p_o.run()
        self.assertEqual(res, set(["IN2P3", "Paris", "CSCS"]))

        p_o.params = {"Site": "", "TargetSE": ["CERN-DST", "CSCS-DST"], "JobType": "User"}
        res = p_o.run()
        self.assertEqual(res, set(["CERN", "CSCS"]))

        # "DataReconstruction" case
        p_o = TaskManagerPlugin("ByJobType", operationsHelper=opsHelperFakeDataReco())

        p_o.params = {"Site": "", "TargetSE": "CERN-DST", "JobType": "DataReconstruction"}
        res = p_o.run()
        self.assertEqual(res, set(["Bologna", "Ferrara", "Paris", "CSCS", "IN2P3", "CERN"]))

        p_o.params = {"Site": "", "TargetSE": "IN2P3-DST", "JobType": "DataReconstruction"}
        res = p_o.run()
        self.assertEqual(res, set(["Bologna", "Paris", "CSCS", "IN2P3"]))

        # "Hospital" case: all sites go to hospital only
        p_o = TaskManagerPlugin("ByJobType", operationsHelper=opsHelperFakeHospital())

        p_o.params = {"Site": "", "TargetSE": "CERN-DST", "JobType": "DataReconstruction"}
        res = p_o.run()
        self.assertEqual(res, set(["Hospital"]))

        p_o.params = {"Site": "", "TargetSE": "IN2P3-DST", "JobType": "DataReconstruction"}
        res = p_o.run()
        self.assertEqual(res, set(["Hospital"]))

        # "Merge" case - 1
        p_o = TaskManagerPlugin("ByJobType", operationsHelper=opsHelperFakeMerge())

        p_o.params = {"Site": "", "TargetSE": "CERN-DST", "JobType": "Merge"}
        res = p_o.run()
        self.assertEqual(res, set(["CERN"]))

        p_o.params = {"Site": "", "TargetSE": "IN2P3-DST", "JobType": "Merge"}
        res = p_o.run()
        self.assertEqual(res, set(["IN2P3"]))

        # "Merge" case - 2
        p_o = TaskManagerPlugin("ByJobType", operationsHelper=opsHelperFakeMerge2())

        p_o.params = {"Site": "", "TargetSE": "CERN-DST", "JobType": "Merge"}
        res = p_o.run()
        self.assertEqual(res, set(["CERN"]))

        p_o.params = {"Site": "", "TargetSE": "IN2P3-DST", "JobType": "Merge"}
        res = p_o.run()
        self.assertEqual(res, set(["IN2P3"]))

        # "MC" case
        p_o = TaskManagerPlugin("ByJobType", operationsHelper=opsHelperFakeMC())

        p_o.params = {"Site": "", "TargetSE": "", "JobType": "MC"}
        res = p_o.run()
        self.assertEqual(res, set(["CERN", "IN2P3", "Bologna", "Ferrara", "Paris", "CSCS", "PAK", "Hospital"]))

        p_o.params = {"Site": "", "TargetSE": "", "JobType": "MC"}
        res = p_o.run()
        self.assertEqual(res, set(["CERN", "IN2P3", "Bologna", "Ferrara", "Paris", "CSCS", "PAK", "Hospital"]))


#############################################################################


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(ClientsTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TaskManagerPluginSuccess))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python
########################################################################
# File :   dirac-version
# Author : Ricardo Graciani
########################################################################
"""
Print version of current DIRAC installation

Usage:
  dirac-version [option]

Example:
  $ dirac-version

"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import argparse

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.parse_known_args()

    print(DIRAC.version)


if __name__ == "__main__":
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import time
import random


from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.ConfigurationSystem.Client.PathFinder import getGatewayURLs
from DIRAC.FrameworkSystem.Client.Logger import gLogger
from DIRAC.Core.Utilities import List
from DIRAC.Core.Utilities.EventDispatcher import gEventDispatcher
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR


def _updateFromRemoteLocation(serviceClient):
    """
    Refresh the configuration
    """
    gLogger.debug("", "Trying to refresh from %s" % serviceClient.serverURL)
    localVersion = gConfigurationData.getVersion()
    retVal = serviceClient.getCompressedDataIfNewer(localVersion)
    if retVal["OK"]:
        dataDict = retVal["Value"]
        newestVersion = dataDict["newestVersion"]
        if localVersion < newestVersion:
            gLogger.debug("New version available", "Updating to version %s..." % newestVersion)
            gConfigurationData.loadRemoteCFGFromCompressedMem(dataDict["data"])
            gLogger.debug("Updated to version %s" % gConfigurationData.getVersion())
            gEventDispatcher.triggerEvent("CSNewVersion", newestVersion, threaded=True)
        return S_OK()
    return retVal


class RefresherBase(object):
    """
    Code factorisation for the refresher
    """

    def __init__(self):
        self._automaticUpdate = False
        self._lastUpdateTime = 0
        self._refreshTime = gConfigurationData.getRefreshTime()
        self._url = False
        self._refreshEnabled = True
        self._timeout = 60
        self._callbacks = {"newVersion": []}
        gEventDispatcher.registerEvent("CSNewVersion")

    def disable(self):
        """
        Disable the refresher and prevent any request to another server
        """
        self._refreshEnabled = False

    def enable(self):
        """
        Enable the refresher and authorize request to another server
        WARNING: It will not activate automatic updates, use autoRefreshAndPublish() for that
        """
        self._refreshEnabled = True
        if self._lastRefreshExpired():
            return self.forceRefresh()
        return S_OK()

    def isEnabled(self):
        """
        Returns if you can use refresher or not, use automaticUpdateEnabled() to know
        if refresh is automatic.
        """
        return self._refreshEnabled

    def addListenerToNewVersionEvent(self, functor):
        gEventDispatcher.addListener("CSNewVersion", functor)

    def _lastRefreshExpired(self):
        """
        Just returns if last refresh must be considered as expired or not
        """
        return time.time() - self._lastUpdateTime >= self._refreshTime

    def forceRefresh(self, fromMaster=False):
        """
        Force refresh
        WARNING: If refresher is disabled, force a refresh will do nothing
        """
        if self._refreshEnabled:
            return self._refresh(fromMaster=fromMaster)
        return S_OK()

    def _refreshAndPublish(self):
        """
        Refresh configuration and publish local updates
        """
        self._lastUpdateTime = time.time()
        gLogger.info("Refreshing from master server")
        sMasterServer = gConfigurationData.getMasterServer()
        if sMasterServer:
            from DIRAC.ConfigurationSystem.Client.ConfigurationClient import ConfigurationClient

            oClient = ConfigurationClient(
                url=sMasterServer,
                timeout=self._timeout,
                useCertificates=gConfigurationData.useServerCertificate(),
                skipCACheck=gConfigurationData.skipCACheck(),
            )
            dRetVal = _updateFromRemoteLocation(oClient)
            if not dRetVal["OK"]:
                gLogger.error("Can't update from master server", dRetVal["Message"])
                return False
            if gConfigurationData.getAutoPublish():
                gLogger.info("Publishing to master server...")
                dRetVal = oClient.publishSlaveServer(self._url)
                if not dRetVal["OK"]:
                    gLogger.error("Can't publish to master server", dRetVal["Message"])
            return True
        else:
            gLogger.warn("No master server is specified in the configuration, trying to get data from other slaves")
            return self._refresh()["OK"]

    def _refresh(self, fromMaster=False):
        """
        Refresh configuration
        """
        self._lastUpdateTime = time.time()
        gLogger.debug("Refreshing configuration...")
        gatewayList = getGatewayURLs("Configuration/Server")
        updatingErrorsList = []
        if gatewayList:
            initialServerList = gatewayList
            gLogger.debug("Using configuration gateway", str(initialServerList[0]))
        elif fromMaster:
            masterServer = gConfigurationData.getMasterServer()
            initialServerList = [masterServer]
            gLogger.debug("Refreshing from master %s" % masterServer)
        else:
            initialServerList = gConfigurationData.getServers()
            gLogger.debug("Refreshing from list %s" % str(initialServerList))

        # If no servers in the initial list, we are supposed to use the local configuration only
        if not initialServerList:
            return S_OK()

        randomServerList = List.randomize(initialServerList)
        gLogger.debug("Randomized server list is %s" % ", ".join(randomServerList))

        for sServer in randomServerList:
            from DIRAC.ConfigurationSystem.Client.ConfigurationClient import ConfigurationClient

            oClient = ConfigurationClient(
                url=sServer,
                useCertificates=gConfigurationData.useServerCertificate(),
                skipCACheck=gConfigurationData.skipCACheck(),
            )
            dRetVal = _updateFromRemoteLocation(oClient)
            if dRetVal["OK"]:
                self._refreshTime = gConfigurationData.getRefreshTime()
                return dRetVal
            else:
                updatingErrorsList.append(dRetVal["Message"])
                gLogger.warn(
                    "Can't update from server", "Error while updating from %s: %s" % (sServer, dRetVal["Message"])
                )
                if dRetVal["Message"].find("Insane environment") > -1:
                    break
        return S_ERROR("Reason(s):\n\t%s" % "\n\t".join(List.uniqueElements(updatingErrorsList)))



########################################################################
# $HeadURL $
# File: CleanReqDBAgent.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2013/05/17 08:31:26
########################################################################
"""Cleaning the RequestDB from obsolete records and kicking assigned requests

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN CleanReqDBAgent
  :end-before: ##END
  :dedent: 2
  :caption: CleanReqDBAgent options

.. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"
# #
# @file CleanReqDBAgent.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2013/05/17 08:32:08
# @brief Definition of CleanReqDBAgent class.

# # imports
import datetime

# # from DIRAC
from DIRAC import S_OK
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.RequestManagementSystem.Client.Request import Request

AGENT_NAME = "RequestManagement/CleanReqDBAgent"

########################################################################


class CleanReqDBAgent(AgentModule):
    """
    .. class:: CleanReqDBAgent

    """

    # # DEL GRACE PERIOD in DAYS
    DEL_GRACE_DAYS = 60
    # # DEL LIMIT
    DEL_LIMIT = 100
    # # KICK PERIOD in HOURS
    KICK_GRACE_HOURS = 1
    # # KICK LIMIT
    KICK_LIMIT = 10000
    # # remove failed requests flag
    DEL_FAILED = False

    # # request client
    __requestClient = None

    def requestClient(self):
        """request client getter"""
        if not self.__requestClient:
            self.__requestClient = ReqClient()
        return self.__requestClient

    def initialize(self):
        """initialization"""
        self.DEL_GRACE_DAYS = self.am_getOption("DeleteGraceDays", self.DEL_GRACE_DAYS)
        self.log.info("Delete grace period = %s days" % self.DEL_GRACE_DAYS)
        self.DEL_LIMIT = self.am_getOption("DeleteLimit", self.DEL_LIMIT)
        self.log.info("Delete limit = %s request/cycle" % self.DEL_LIMIT)
        self.DEL_FAILED = self.am_getOption("DeleteFailed", self.DEL_FAILED)
        self.log.info("Delete failed requests: %s" % {True: "yes", False: "no"}[self.DEL_FAILED])

        self.KICK_GRACE_HOURS = self.am_getOption("KickGraceHours", self.KICK_GRACE_HOURS)
        self.log.info("Kick assigned requests period = %s hours" % self.KICK_GRACE_HOURS)
        self.KICK_LIMIT = self.am_getOption("KickLimit", self.KICK_LIMIT)
        self.log.info("Kick limit = %s request/cycle" % self.KICK_LIMIT)

        # # gMonitor stuff
        gMonitor.registerActivity(
            "DeletedRequests", "Deleted finished requests", "CleanReqDBAgent", "Requests/min", gMonitor.OP_SUM
        )
        gMonitor.registerActivity(
            "KickedRequests", "Assigned requests kicked", "CleanReqDBAgent", "Requests/min", gMonitor.OP_SUM
        )
        return S_OK()

    def execute(self):
        """execution in one cycle"""

        now = datetime.datetime.utcnow()
        kickTime = now - datetime.timedelta(hours=self.KICK_GRACE_HOURS)
        rmTime = now - datetime.timedelta(days=self.DEL_GRACE_DAYS)

        # # kick
        statusList = ["Assigned"]
        requestIDsList = self.requestClient().getRequestIDsList(statusList, self.KICK_LIMIT)
        if not requestIDsList["OK"]:
            self.log.error("execute: %s" % requestIDsList["Message"])
            return requestIDsList

        requestIDsList = requestIDsList["Value"]
        kicked = 0
        for requestID, status, lastUpdate in requestIDsList:
            reqStatus = self.requestClient().getRequestStatus(requestID)
            if not reqStatus["OK"]:
                self.log.error(("execute: unable to get request status", reqStatus["Message"]))
                continue
            status = reqStatus["Value"]
            if lastUpdate < kickTime and status == "Assigned":
                getRequest = self.requestClient().peekRequest(requestID)
                if not getRequest["OK"]:
                    self.log.error("execute: unable to read request '%s': %s" % (requestID, getRequest["Message"]))
                    continue
                getRequest = getRequest["Value"]
                if getRequest and getRequest.LastUpdate < kickTime:
                    self.log.info(
                        "execute: kick assigned request (%s/'%s') in status %s"
                        % (requestID, getRequest.RequestName, getRequest.Status)
                    )
                    putRequest = self.requestClient().putRequest(getRequest)
                    if not putRequest["OK"]:
                        self.log.error(
                            "execute: unable to put request (%s/'%s'): %s"
                            % (requestID, getRequest.RequestName, putRequest["Message"])
                        )
                        continue
                    else:
                        self.log.verbose("Kicked request %d" % putRequest["Value"])
                    kicked += 1

        # # delete
        statusList = ["Done", "Failed", "Canceled"] if self.DEL_FAILED else ["Done"]
        requestIDsList = self.requestClient().getRequestIDsList(statusList, self.DEL_LIMIT)
        if not requestIDsList["OK"]:
            self.log.error("execute: %s" % requestIDsList["Message"])
            return requestIDsList

        requestIDsList = requestIDsList["Value"]
        deleted = 0
        for requestID, status, lastUpdate in requestIDsList:
            if lastUpdate < rmTime:
                self.log.info("execute: deleting request '%s' with status %s" % (requestID, status))
                delRequest = self.requestClient().deleteRequest(requestID)
                if not delRequest["OK"]:
                    self.log.error("execute: unable to delete request '%s': %s" % (requestID, delRequest["Message"]))
                    continue
                deleted += 1

        gMonitor.addMark("KickedRequests", kicked)
        gMonitor.addMark("DeletedRequests", deleted)
        self.log.info("execute: kicked assigned requests = %s" % kicked)
        self.log.info("execute: deleted finished requests = %s" % deleted)
        return S_OK()



""" Soft implementation of a Direct Acyclic Graph (DAG)

    Nota Bene: It is NOT fully checked if valid (i.e. some cycle can be introduced)!
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
from DIRAC import gLogger

__RCSID__ = "$Id $"


class DAG(object):
    """a Direct Acyclic Graph (DAG)

    Represented as a dictionary whose keys are nodes, and values are sets holiding their dependencies
    """

    def __init__(self):
        """Defines graph variable holding the dag representation"""
        self.graph = {}

    def addNode(self, node):
        """add a node to graph

        Args:
          node (object): Any type of object - if not hashable, it will be converted to a frozenset
        """
        node = checkNode(node)
        if node not in self.graph:
            self.graph[node] = set()

    def addEdge(self, fromNode, toNode):
        """add an edge (checks if both nodes exist)

        Args:
          fromNode (object)
          toNode (object)
        """
        fromNode = checkNode(fromNode)
        toNode = checkNode(toNode)

        if fromNode not in self.graph:
            gLogger.error("Missing node from where the edge start")
            return
        if toNode not in self.graph:
            gLogger.error("Missing node to where the edge lands")
            return

        for node, toNodes in self.graph.items():
            # This is clearly not enough to assure that it's really acyclic...
            if toNode == node and fromNode in toNodes:
                gLogger.error("Can't insert this edge")
                return
        self.graph[fromNode].add(toNode)

    def getIndexNodes(self):
        """Return a list of index nodes"""
        notIndexNodes = set()
        for depNodes in self.graph.values():
            [notIndexNodes.add(depNode) for depNode in depNodes]
        indexNodes = list(set(self.graph.keys()) - notIndexNodes)
        return [unHashNode(inu) for inu in indexNodes]

    def getList(self):
        """Returns a list out of the DAG, if possible"""
        cDAG = copy.deepcopy(self)

        l = []
        while True:
            try:
                indexNodes = cDAG.getIndexNodes()
                if len(indexNodes) != 1:
                    gLogger.warn("The DAG is not sequential")
                    break
                ind = checkNode(indexNodes[0])
                del cDAG.graph[ind]
                l.append(unHashNode(ind))
            except KeyError:
                break
            except IndexError:
                break

        return l


def checkNode(node):
    """Returns a hashable version of node"""
    try:
        node.__hash__()
        return node
    except TypeError:  # nodeName is not hashable, so it can't be a key in the graph (which is a dictionary)
        return makeFrozenSet(node)


def unHashNode(node):
    """Returns a dict or list, if node is frozenset"""
    if isinstance(node, frozenset):
        try:
            return dict(node)  # Is it a dictionary?
        except TypeError:
            return list(node)
        except ValueError:
            return list(node)
    else:
        return node


def makeFrozenSet(ob):
    """
    Makes a hash from a dictionary, list, tuple or set to any level, that contains
    only other hashable types (including any lists, tuples, sets, and dictionaries).
    """
    if isinstance(ob, (set, tuple, list)):
        return frozenset([makeFrozenSet(e) for e in ob])

    elif not isinstance(ob, dict):
        return ob

    new_ob = copy.deepcopy(ob)
    for obk, obv in new_ob.items():
        new_ob[obk] = makeFrozenSet(obv)

    return frozenset(sorted(new_ob.items()))



#!/bin/env python
"""
tests for SSHComputingElement module
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import shutil
import subprocess32 as subprocess
import shlex
import pytest

import DIRAC

from DIRAC.Resources.Computing.SSHComputingElement import SSHComputingElement
from DIRAC.Resources.Computing.BatchSystems.executeBatch import executeBatchContent


@pytest.mark.parametrize("batchSystem", ["Condor", "GE", "Host", "LSF", "OAR", "SLURM", "Torque"])
def test_generateControlScript(batchSystem):
    """Test that the control script generated by the merging operation
    between a BatchSystem and executeBatch.py is:
    * complete: contains the content of both files
    * executable and doesn't raise any syntax error.

    Example: it may check that a __future__ import is not misplaced in the script due to the
    merging of the files.
    """

    ce = SSHComputingElement("Test_SSHCE")
    # Change the batch system file used during the control script generation
    ce.loadBatchSystem(batchSystem)
    # Get the local control script
    result = ce._generateControlScript()
    assert result["OK"] is True

    source = result["Value"]
    dest = "execute_batch.py"

    # Simulate operation done by the scpCall method

    # Copy the local control script into the "remote" control script
    # As the source can be composed of multiple files, we have to copy the content of each file
    sources = source.split(" ")
    with open(dest, "wb") as dst:
        for sourceFile in sources:
            with open(sourceFile, "rb") as src:
                shutil.copyfileobj(src, dst)

    # Test that the control script is complete
    with open(dest, "r") as dst:
        dataDest = dst.read()

    batchSystemDir = os.path.join(os.path.dirname(DIRAC.__file__), "Resources", "Computing", "BatchSystems")
    batchSystemScript = os.path.join(batchSystemDir, "%s.py" % batchSystem)
    with open(batchSystemScript, "r") as bsc:
        dataBatchSystemScript = bsc.read()

    assert executeBatchContent in dataDest
    assert dataBatchSystemScript in dataDest

    # Test the execution of the remote control script
    cmd = "python -m py_compile %s" % dest
    args = shlex.split(cmd)
    process = subprocess.Popen(args, universal_newlines=True)
    process.communicate()
    assert process.returncode == 0

    # Delete the control script and the .pyc file associated
    os.remove(source)
    os.remove(dest)
    if os.path.isfile("%sc" % dest):
        os.remove("%sc" % dest)



"""
The Job Wrapper Class is instantiated with arguments tailored for running
a particular job. The JobWrapper starts a thread for execution of the job
and a Watchdog Agent that can monitor its progress.


.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN JobWrapper
  :end-before: ##END
  :caption: JobWrapper options

"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import os
import stat
import re
import sys
import time
import shutil
import threading
import tarfile
import glob
import json
import distutils.spawn
import six

from six.moves.urllib.parse import unquote as urlunquote

import DIRAC
from DIRAC import S_OK, S_ERROR, gConfig, gLogger
from DIRAC.AccountingSystem.Client.Types.Job import Job as AccountingJob
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Utilities import List
from DIRAC.Core.Utilities import DEncode
from DIRAC.Core.Utilities import Time
from DIRAC.Core.Utilities.SiteSEMapping import getSEsForSite
from DIRAC.Core.Utilities.ModuleFactory import ModuleFactory
from DIRAC.Core.Utilities.Subprocess import systemCall
from DIRAC.Core.Utilities.Subprocess import Subprocess
from DIRAC.Core.Utilities.File import getGlobbedTotalSize, getGlobbedFiles
from DIRAC.Core.Utilities.Version import getCurrentVersion
from DIRAC.Core.Utilities.Adler import fileAdler
from DIRAC.ConfigurationSystem.Client.PathFinder import getSystemSection
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOForGroup
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.DataManagementSystem.Client.FailoverTransfer import FailoverTransfer
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import resolveSEGroup
from DIRAC.Resources.Catalog.PoolXMLFile import getGUID
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.RequestManagementSystem.private.RequestValidator import RequestValidator
from DIRAC.WorkloadManagementSystem.JobWrapper.Watchdog import Watchdog
from DIRAC.WorkloadManagementSystem.Client.JobStateUpdateClient import JobStateUpdateClient
from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import JobMonitoringClient
from DIRAC.WorkloadManagementSystem.Client.JobManagerClient import JobManagerClient
from DIRAC.WorkloadManagementSystem.Client.SandboxStoreClient import SandboxStoreClient
from DIRAC.WorkloadManagementSystem.Client.JobReport import JobReport
from DIRAC.WorkloadManagementSystem.Client import JobStatus
from DIRAC.WorkloadManagementSystem.Client import JobMinorStatus

EXECUTION_RESULT = {}


class JobWrapper(object):
    """The only user of the JobWrapper is the JobWrapperTemplate"""

    #############################################################################
    def __init__(self, jobID=None, jobReport=None):
        """Standard constructor"""
        self.initialTiming = os.times()
        self.section = os.path.join(getSystemSection("WorkloadManagement/JobWrapper"), "JobWrapper")
        # Create the accounting report
        self.accountingReport = AccountingJob()
        # Initialize for accounting
        self.wmsMajorStatus = "unknown"
        self.wmsMinorStatus = "unknown"
        # Set now as start time
        self.accountingReport.setStartTime()
        if not jobID:
            self.jobID = 0
        else:
            self.jobID = jobID
        self.siteName = DIRAC.siteName()
        if jobReport:
            self.jobReport = jobReport
        else:
            self.jobReport = JobReport(self.jobID, "JobWrapper@%s" % self.siteName)
        self.failoverTransfer = FailoverTransfer()

        self.log = gLogger.getSubLogger("JobWrapper[%s]" % self.jobID)
        self.log.showHeaders(True)

        # self.root is the path the Wrapper is running at
        self.root = os.getcwd()
        # self.localSiteRoot is the path where the local DIRAC installation used to run the payload
        # is taken from
        self.localSiteRoot = gConfig.getValue("/LocalSite/Root", DIRAC.rootPath)
        # FIXME: Why do we need to load any .cfg file here????
        self.__loadLocalCFGFiles(self.localSiteRoot)
        result = getCurrentVersion()
        if result["OK"]:
            self.diracVersion = result["Value"]
        else:
            self.diracVersion = "DIRAC version %s" % DIRAC.version
        self.maxPeekLines = gConfig.getValue(self.section + "/MaxJobPeekLines", 20)
        if self.maxPeekLines < 0:
            self.maxPeekLines = 0
        self.defaultCPUTime = gConfig.getValue(self.section + "/DefaultCPUTime", 600)
        self.defaultOutputFile = gConfig.getValue(self.section + "/DefaultOutputFile", "std.out")
        self.defaultErrorFile = gConfig.getValue(self.section + "/DefaultErrorFile", "std.err")
        self.diskSE = gConfig.getValue(self.section + "/DiskSE", ["-disk", "-DST", "-USER"])
        self.tapeSE = gConfig.getValue(self.section + "/TapeSE", ["-tape", "-RDST", "-RAW"])
        self.failoverRequestDelay = gConfig.getValue(self.section + "/FailoverRequestDelay", 45)
        self.sandboxSizeLimit = gConfig.getValue(self.section + "/OutputSandboxLimit", 1024 * 1024 * 10)
        self.cleanUpFlag = gConfig.getValue(self.section + "/CleanUpFlag", True)
        self.boincUserID = gConfig.getValue("/LocalSite/BoincUserID", 0)
        self.pilotRef = gConfig.getValue("/LocalSite/PilotReference", "Unknown")
        self.cpuNormalizationFactor = gConfig.getValue("/LocalSite/CPUNormalizationFactor", 0.0)
        self.bufferLimit = gConfig.getValue(self.section + "/BufferLimit", 10485760)
        self.defaultOutputSE = resolveSEGroup(gConfig.getValue("/Resources/StorageElementGroups/SE-USER", []))
        self.defaultCatalog = gConfig.getValue(self.section + "/DefaultCatalog", [])
        self.masterCatalogOnlyFlag = gConfig.getValue(self.section + "/MasterCatalogOnlyFlag", True)
        self.defaultFailoverSE = resolveSEGroup(gConfig.getValue("/Resources/StorageElementGroups/Tier1-Failover", []))
        self.defaultOutputPath = ""
        self.retryUpload = gConfig.getValue(self.section + "/RetryUpload", False)
        self.dm = DataManager()
        self.fc = FileCatalog()
        self.log.verbose("===========================================================================")
        self.log.verbose("Version %s" % (__RCSID__))
        self.log.verbose(self.diracVersion)
        self.currentPID = os.getpid()
        self.log.verbose("Job Wrapper started under PID: %s" % self.currentPID)
        # Define a new process group for the job wrapper
        self.parentPGID = os.getpgid(self.currentPID)
        self.log.verbose("Job Wrapper parent process group ID: %s" % self.parentPGID)
        os.setpgid(self.currentPID, self.currentPID)
        self.currentPGID = os.getpgid(self.currentPID)
        self.log.verbose("Job Wrapper process group ID: %s" % self.currentPGID)
        self.log.verbose("==========================================================================")
        self.log.verbose("sys.path is: \n%s" % "\n".join(sys.path))
        self.log.verbose("==========================================================================")
        if "PYTHONPATH" not in os.environ:
            self.log.verbose("PYTHONPATH is: null")
        else:
            pypath = os.environ["PYTHONPATH"]
            self.log.verbose("PYTHONPATH is: \n%s" % "\n".join(pypath.split(":")))
            self.log.verbose("==========================================================================")
        if "LD_LIBRARY_PATH_SAVE" in os.environ:
            if "LD_LIBRARY_PATH" in os.environ:
                os.environ["LD_LIBRARY_PATH"] += ":" + os.environ["LD_LIBRARY_PATH_SAVE"]
            else:
                os.environ["LD_LIBRARY_PATH"] = os.environ["LD_LIBRARY_PATH_SAVE"]

        if "LD_LIBRARY_PATH" not in os.environ:
            self.log.verbose("LD_LIBRARY_PATH is: null")
        else:
            ldpath = os.environ["LD_LIBRARY_PATH"]
            self.log.verbose("LD_LIBRARY_PATH is: \n%s" % "\n".join(ldpath.split(":")))
            self.log.verbose("==========================================================================")
        if not self.cleanUpFlag:
            self.log.verbose("CleanUp Flag is disabled by configuration")
        # Failure flag
        self.failedFlag = True
        # Set defaults for some global parameters to be defined for the accounting report
        self.owner = "unknown"
        self.jobGroup = "unknown"
        self.jobType = "unknown"
        self.processingType = "unknown"
        self.userGroup = "unknown"
        self.jobClass = "unknown"
        self.inputDataFiles = 0
        self.outputDataFiles = 0
        self.inputDataSize = 0
        self.inputSandboxSize = 0
        self.outputSandboxSize = 0
        self.outputDataSize = 0
        self.processedEvents = 0
        self.numberOfProcessors = 1
        self.jobAccountingSent = False

        self.jobArgs = {}
        self.optArgs = {}
        self.ceArgs = {}

    #############################################################################
    def initialize(self, arguments):
        """Initializes parameters and environment for job."""
        self.__report(status=JobStatus.RUNNING, minorStatus=JobMinorStatus.JOB_INITIALIZATION)
        self.log.info("Starting Job Wrapper Initialization for Job", self.jobID)
        self.jobArgs = arguments["Job"]
        self.log.verbose(self.jobArgs)
        self.ceArgs = arguments["CE"]
        self.log.verbose(self.ceArgs)
        self.__setInitialJobParameters()
        self.optArgs = arguments.get("Optimizer", {})
        # Fill some parameters for the accounting report
        self.owner = self.jobArgs.get("Owner", self.owner)
        self.jobGroup = self.jobArgs.get("JobGroup", self.jobGroup)
        self.jobType = self.jobArgs.get("JobType", self.jobType)
        dataParam = self.jobArgs.get("InputData", [])
        if dataParam and not isinstance(dataParam, list):
            dataParam = [dataParam]
        self.inputDataFiles = len(dataParam)
        dataParam = self.jobArgs.get("OutputData", [])
        if dataParam and not isinstance(dataParam, list):
            dataParam = [dataParam]
        self.outputDataFiles = len(dataParam)
        self.processingType = self.jobArgs.get("ProcessingType", self.processingType)
        self.userGroup = self.jobArgs.get("OwnerGroup", self.userGroup)
        self.jobClass = self.jobArgs.get("JobSplitType", self.jobClass)

        # Prepare the working directory, cd to there, and copying eventual extra arguments in it
        if self.jobID:
            if os.path.exists(str(self.jobID)):
                shutil.rmtree(str(self.jobID))
            os.mkdir(str(self.jobID))
            os.chdir(str(self.jobID))
            extraOpts = self.jobArgs.get("ExtraOptions", "")
            if extraOpts and "dirac-jobexec" in self.jobArgs.get("Executable", "").strip():
                if os.path.exists("%s/%s" % (self.root, extraOpts)):
                    shutil.copyfile("%s/%s" % (self.root, extraOpts), extraOpts)
                self.__loadLocalCFGFiles(self.localSiteRoot)

        else:
            self.log.info("JobID is not defined, running in current directory")

        with open("job.info", "w") as infoFile:
            infoFile.write(self.__dictAsInfoString(self.jobArgs, "/Job"))

        self.log.debug("Environment used")
        self.log.debug("================")
        self.log.debug(json.dumps(dict(os.environ), indent=4))

    #############################################################################
    def __setInitialJobParameters(self):
        """Sets some initial job parameters"""
        parameters = []
        parameters.append(("Pilot_Reference", self.ceArgs.get("PilotReference", self.pilotRef)))
        if "LocalSE" in self.ceArgs:
            parameters.append(("AgentLocalSE", ",".join(self.ceArgs["LocalSE"])))
        if "CPUNormalizationFactor" in self.ceArgs:
            parameters.append(("CPUNormalizationFactor", self.ceArgs["CPUNormalizationFactor"]))
        if self.boincUserID:
            parameters.append(("BoincUserID", self.boincUserID))

        parameters.append(("PilotAgent", self.diracVersion))
        parameters.append(("JobWrapperPID", self.currentPID))
        return self.__setJobParamList(parameters)

    #############################################################################
    def __loadLocalCFGFiles(self, localRoot):
        """Loads any extra CFG files residing in the local DIRAC site root."""
        files = os.listdir(localRoot)
        self.log.debug("Checking directory %s for *.cfg files" % localRoot)
        for localFile in files:
            if re.search(".cfg$", localFile):
                gConfig.loadFile("%s/%s" % (localRoot, localFile))
                self.log.verbose("Found local .cfg file '%s'" % localFile)

    #############################################################################
    def __dictAsInfoString(self, dData, infoString="", currentBase=""):
        for key in dData:
            value = dData[key]
            if isinstance(value, dict):
                infoString = self.__dictAsInfoString(value, infoString, "%s/%s" % (currentBase, key))
            elif isinstance(value, (list, tuple)):
                if value and value[0] == "[":
                    infoString += "%s/%s = %s\n" % (currentBase, key, " ".join(value))
                else:
                    infoString += "%s/%s = %s\n" % (currentBase, key, ", ".join(value))
            else:
                infoString += "%s/%s = %s\n" % (currentBase, key, str(value))

        return infoString

    #############################################################################
    def execute(self):
        """The main execution method of the Job Wrapper"""
        self.log.info("Job Wrapper is starting execution phase for job %s" % (self.jobID))
        os.environ["DIRACJOBID"] = str(self.jobID)
        os.environ["DIRACROOT"] = self.localSiteRoot
        self.log.verbose("DIRACROOT = %s" % (self.localSiteRoot))
        os.environ["DIRACSITE"] = DIRAC.siteName()
        self.log.verbose("DIRACSITE = %s" % (DIRAC.siteName()))

        os.environ["DIRAC_PROCESSORS"] = str(self.ceArgs.get("Processors", 1))
        self.log.verbose("DIRAC_PROCESSORS = %s" % (self.ceArgs.get("Processors", 1)))

        os.environ["DIRAC_WHOLENODE"] = str(self.ceArgs.get("WholeNode", False))
        self.log.verbose("DIRAC_WHOLENODE = %s" % (self.ceArgs.get("WholeNode", False)))

        errorFile = self.jobArgs.get("StdError", self.defaultErrorFile)
        outputFile = self.jobArgs.get("StdOutput", self.defaultOutputFile)

        if "CPUTime" in self.jobArgs:
            jobCPUTime = int(self.jobArgs["CPUTime"])
        else:
            self.log.info(
                "Job has no CPU time limit specified, ",
                "applying default of %s to %s" % (self.defaultCPUTime, self.jobID),
            )
            jobCPUTime = self.defaultCPUTime
        self.numberOfProcessors = int(self.jobArgs.get("NumberOfProcessors", 1))

        jobMemory = 0.0
        if "Memory" in self.jobArgs:
            # Job specifies memory in GB, internally use KB
            jobMemory = int(self.jobArgs["Memory"]) * 1024.0 * 1024.0

        if "Executable" in self.jobArgs:
            executable = self.jobArgs[
                "Executable"
            ].strip()  # This is normally dirac-jobexec script, but not necessarily
        else:
            msg = "Job %s has no specified executable" % (self.jobID)
            self.log.warn(msg)
            return S_ERROR(msg)

        # In case the executable is dirac-jobexec,
        # the argument should include the jobDescription.xml file
        jobArguments = self.jobArgs.get("Arguments", "")

        # This is a workaround for Python 2 style installations
        if six.PY3 and executable == "$DIRACROOT/scripts/dirac-jobexec":
            self.log.warn(
                'Replaced job executable "$DIRACROOT/scripts/dirac-jobexec" with '
                '"dirac-jobexec". Please fix your submission script!'
            )
            executable = "dirac-jobexec"

        executable = os.path.expandvars(executable)
        exeThread = None
        spObject = None

        if re.search("DIRACROOT", executable):
            executable = executable.replace("$DIRACROOT", self.localSiteRoot)
            self.log.verbose("Replaced $DIRACROOT for executable as %s" % (self.localSiteRoot))

        # Try to find the executable on PATH
        if "/" not in executable:
            # Returns None if the executable is not found so use "or" to leave it unchanged
            executable = distutils.spawn.find_executable(executable) or executable

        # Make the full path since . is not always in the PATH
        executable = os.path.abspath(executable)
        if not os.access(executable, os.X_OK):
            try:
                os.chmod(executable, stat.S_IRWXU | stat.S_IRWXG | stat.S_IROTH | stat.S_IXOTH)
            except OSError:
                self.log.warn("Failed to change mode to 775 for the executable", executable)

        exeEnv = dict(os.environ)
        if "ExecutionEnvironment" in self.jobArgs:
            self.log.verbose("Adding variables to execution environment")
            variableList = self.jobArgs["ExecutionEnvironment"]
            if isinstance(variableList, six.string_types):
                variableList = [variableList]
            for var in variableList:
                nameEnv = var.split("=")[0]
                valEnv = urlunquote(var.split("=")[1])
                exeEnv[nameEnv] = valEnv
                self.log.verbose("%s = %s" % (nameEnv, valEnv))

        if os.path.exists(executable):
            # it's in fact not yet running: it will be in few lines
            self.__report(status=JobStatus.RUNNING, minorStatus=JobMinorStatus.APPLICATION, sendFlag=True)
            spObject = Subprocess(timeout=False, bufferLimit=int(self.bufferLimit))
            command = executable
            if jobArguments:
                command += " " + jobArguments
            self.log.verbose("Execution command: %s" % (command))
            maxPeekLines = self.maxPeekLines
            exeThread = ExecutionThread(spObject, command, maxPeekLines, outputFile, errorFile, exeEnv)
            exeThread.start()
            payloadPID = None
            for seconds in range(5, 40, 5):
                time.sleep(seconds)
                payloadPID = spObject.getChildPID()
                if payloadPID:
                    self.__setJobParam("PayloadPID", payloadPID)
                    break
            if not payloadPID:
                return S_ERROR("Payload process could not start after 140 seconds")
        else:
            self.__report(status=JobStatus.FAILED, minorStatus=JobMinorStatus.APP_NOT_FOUND, sendFlag=True)
            return S_ERROR("Path to executable %s not found" % (executable))

        watchdog = Watchdog(
            pid=self.currentPID,
            exeThread=exeThread,
            spObject=spObject,
            jobCPUTime=jobCPUTime,
            memoryLimit=jobMemory,
            processors=self.numberOfProcessors,
            jobArgs=self.jobArgs,
        )

        self.log.verbose("Initializing Watchdog instance")
        watchdog.initialize()
        self.log.verbose("Calibrating Watchdog instance")
        watchdog.calibrate()
        # do not kill Test jobs by CPU time
        if self.jobArgs.get("JobType", "") == "Test":
            watchdog.testCPUConsumed = False

        if "DisableCPUCheck" in self.jobArgs:
            watchdog.testCPUConsumed = False

        if exeThread.is_alive():
            self.log.info("Application thread is started in Job Wrapper")
            watchdog.run()
        else:
            self.log.warn("Application thread stopped very quickly...")

        if exeThread.is_alive():
            self.log.warn("Watchdog exited before completion of execution thread")
            while exeThread.is_alive():
                time.sleep(5)

        outputs = None
        if "Thread" in EXECUTION_RESULT:
            threadResult = EXECUTION_RESULT["Thread"]
            if not threadResult["OK"]:
                self.log.error("Failed to execute the payload", threadResult["Message"])

                self.__report(status=JobStatus.FAILED, minorStatus=JobMinorStatus.APP_THREAD_FAILED, sendFlag=True)
                if "Value" in threadResult:
                    outs = threadResult["Value"]
                if outs:
                    self.__setJobParam("ApplicationError", outs[0], sendFlag=True)
                else:
                    self.__setJobParam("ApplicationError", "None reported", sendFlag=True)
            else:
                outputs = threadResult["Value"]
        else:  # if the execution thread didn't complete
            self.log.error("Application thread did not complete")
            self.__report(status=JobStatus.FAILED, minorStatus=JobMinorStatus.APP_THREAD_NOT_COMPLETE, sendFlag=True)
            self.__setJobParam("ApplicationError", JobMinorStatus.APP_THREAD_NOT_COMPLETE, sendFlag=True)
            return S_ERROR("No outputs generated from job execution")

        if "CPU" in EXECUTION_RESULT:
            cpuString = " ".join(["%.2f" % x for x in EXECUTION_RESULT["CPU"]])
            self.log.info("EXECUTION_RESULT[CPU] in JobWrapper execute", cpuString)

        if watchdog.checkError:
            # In this case, the Watchdog has killed the Payload and the ExecutionThread can not get the CPU statistics
            # os.times only reports for waited children
            # Take the CPU from the last value recorded by the Watchdog
            self.__report(status=JobStatus.FAILED, minorStatus=watchdog.checkError, sendFlag=True)
            if "CPU" in EXECUTION_RESULT:
                if "LastUpdateCPU(s)" in watchdog.currentStats:
                    EXECUTION_RESULT["CPU"][0] = 0
                    EXECUTION_RESULT["CPU"][0] = 0
                    EXECUTION_RESULT["CPU"][0] = 0
                    EXECUTION_RESULT["CPU"][0] = watchdog.currentStats["LastUpdateCPU(s)"]

        if watchdog.currentStats:
            self.log.info(
                "Statistics collected by the Watchdog:\n ",
                "\n  ".join(["%s: %s" % items for items in watchdog.currentStats.items()]),
            )  # can be an iterator
        if outputs:
            status = threadResult["Value"][0]  # the status of the payload execution
            # Send final heartbeat of a configurable number of lines here
            self.log.verbose("Sending final application standard output heartbeat")
            self.__sendFinalStdOut(exeThread)
            self.log.verbose("Execution thread status = %s" % (status))

            if not watchdog.checkError and not status:
                self.failedFlag = False
                self.__report(status=JobStatus.COMPLETING, minorStatus=JobMinorStatus.APP_SUCCESS, sendFlag=True)
            elif not watchdog.checkError:
                self.__report(status=JobStatus.COMPLETING, minorStatus=JobMinorStatus.APP_ERRORS, sendFlag=True)
                if status in (DErrno.EWMSRESC, DErrno.EWMSRESC & 255):  # the status will be truncated to 0xDE (222)
                    self.log.verbose("job will be rescheduled")
                    self.__report(
                        status=JobStatus.COMPLETING, minorStatus=JobMinorStatus.GOING_RESCHEDULE, sendFlag=True
                    )
                    return S_ERROR(DErrno.EWMSRESC, "Job will be rescheduled")

        else:
            return S_ERROR("No outputs generated from job execution")

        self.log.info("Checking directory contents after execution:")
        res = systemCall(5, ["ls", "-al"])
        if not res["OK"]:
            self.log.error("Failed to list the current directory", res["Message"])
        elif res["Value"][0]:
            self.log.error("Failed to list the current directory", res["Value"][2])
        else:
            # no timeout and exit code is 0
            self.log.info(res["Value"][1])

        return S_OK()

    #############################################################################
    def __sendFinalStdOut(self, exeThread):
        """After the Watchdog process has finished, this function sends a final
        report to be presented in the StdOut in the web page via the heartbeat
        mechanism.
        """
        cpuConsumed = self.__getCPU()["Value"]
        self.log.info("Total CPU Consumed is: %s" % cpuConsumed[1])
        self.__setJobParam("TotalCPUTime(s)", cpuConsumed[0])
        normCPU = cpuConsumed[0] * self.cpuNormalizationFactor
        self.__setJobParam("NormCPUTime(s)", normCPU)
        if self.cpuNormalizationFactor:
            self.log.info("Normalized CPU Consumed is:", normCPU)

        result = exeThread.getOutput(self.maxPeekLines)
        if not result["OK"]:
            lines = 0
            appStdOut = ""
        else:
            lines = len(result["Value"])
            appStdOut = "\n".join(result["Value"])

        header = "Last %s lines of application output from JobWrapper on %s :" % (lines, Time.toString())
        border = "=" * len(header)

        cpuTotal = "CPU Total: %s (h:m:s)" % cpuConsumed[1]
        cpuTotal += " Normalized CPU Total %.1f s @ HEP'06" % normCPU
        header = "\n%s\n%s\n%s\n%s\n" % (border, header, cpuTotal, border)
        appStdOut = header + appStdOut
        self.log.info(appStdOut)
        heartBeatDict = {}
        staticParamDict = {"StandardOutput": appStdOut}
        if self.jobID:
            result = JobStateUpdateClient().sendHeartBeat(self.jobID, heartBeatDict, staticParamDict)
            if not result["OK"]:
                self.log.error("Problem sending final heartbeat from JobWrapper", result["Message"])

        return

    #############################################################################
    def __getCPU(self):
        """Uses os.times() to get CPU time and returns HH:MM:SS after conversion."""
        # TODO: normalize CPU consumed via scale factor
        cpuString = " ".join(["%.2f" % x for x in EXECUTION_RESULT["CPU"]])
        self.log.info("EXECUTION_RESULT[CPU] in __getCPU", cpuString)
        utime, stime, cutime, cstime, _elapsed = EXECUTION_RESULT["CPU"]
        cpuTime = utime + stime + cutime + cstime
        self.log.verbose("Total CPU time consumed = %s" % (cpuTime))
        result = self.__getCPUHMS(cpuTime)
        return result

    #############################################################################
    def __getCPUHMS(self, cpuTime):
        mins, secs = divmod(cpuTime, 60)
        hours, mins = divmod(mins, 60)
        humanTime = "%02d:%02d:%02d" % (hours, mins, secs)
        self.log.verbose("Human readable CPU time is: %s" % humanTime)
        return S_OK((cpuTime, humanTime))

    #############################################################################
    def resolveInputData(self):
        """Input data is resolved here using a VO specific plugin module."""
        self.__report(status=JobStatus.RUNNING, minorStatus=JobMinorStatus.INPUT_DATA_RESOLUTION, sendFlag=True)

        # What is this input data? - and exit if there's no input
        inputData = self.jobArgs["InputData"]
        if not inputData:
            msg = "Job Wrapper cannot resolve local replicas of input data with null job input data parameter "
            self.log.error(msg)
            return S_ERROR(msg)
        else:
            if isinstance(inputData, six.string_types):
                inputData = [inputData]
            lfns = [fname.replace("LFN:", "") for fname in inputData]
            self.log.verbose("Job input data requirement is \n%s" % ",\n".join(lfns))

        # Does this site have local SEs? - not failing if it doesn't
        if "LocalSE" in self.ceArgs:
            localSEList = self.ceArgs["LocalSE"]
        else:
            localSEList = gConfig.getValue("/LocalSite/LocalSE", [])

        if not localSEList:
            self.log.warn("Job has input data requirement but no site LocalSE defined")
        else:
            if isinstance(localSEList, six.string_types):
                localSEList = List.fromChar(localSEList)
            self.log.info("Site has the following local SEs: %s" % ", ".join(localSEList))

        # How to get this data?
        if "InputDataModule" not in self.jobArgs:
            self.log.warn("Job has no input data resolution module specified, using the default one")
            inputDataPolicy = "DIRAC.WorkloadManagementSystem.Client.InputDataResolution"
        else:
            inputDataPolicy = self.jobArgs["InputDataModule"]

        self.log.verbose("Job input data resolution policy module is %s" % (inputDataPolicy))

        # Now doing the real stuff
        optReplicas = {}
        if self.optArgs:
            try:
                optDict, _length = DEncode.decode(self.optArgs["InputData"])
                optReplicas = optDict["Value"]
                self.log.info("Found optimizer catalog result")
                self.log.verbose(optReplicas)
            except Exception as x:
                self.log.warn(str(x))
                self.log.warn("Optimizer information could not be converted to a dictionary will call catalog directly")

        result = self.__checkFileCatalog(lfns, optReplicas)
        if not result["OK"]:
            self.log.info("Could not obtain replica information from Optimizer File Catalog information")
            self.log.warn(result)
            result = self.__checkFileCatalog(lfns)
            if not result["OK"]:
                self.log.warn("Could not obtain replica information from File Catalog directly")
                self.log.warn(result)
                return S_ERROR(result["Message"])
            else:
                resolvedData = result
        else:
            resolvedData = result

        # add input data size to accounting report (since resolution successful)
        for lfn, mdata in resolvedData["Value"]["Successful"].items():  # can be an iterator
            if "Size" in mdata:
                lfnSize = mdata["Size"]
                if not isinstance(lfnSize, six.integer_types):
                    try:
                        lfnSize = int(lfnSize)
                    except ValueError:
                        lfnSize = 0
                        self.log.info("File size for LFN was not an integer, setting size to 0", lfn)
                self.inputDataSize += lfnSize

        configDict = {
            "JobID": self.jobID,
            "LocalSEList": localSEList,
            "DiskSEList": self.diskSE,
            "TapeSEList": self.tapeSE,
        }
        self.log.info(configDict)
        argumentsDict = {
            "FileCatalog": resolvedData,
            "Configuration": configDict,
            "InputData": lfns,
            "Job": self.jobArgs,
        }
        self.log.info(argumentsDict)
        moduleFactory = ModuleFactory()
        self.log.verbose("Now starting execution of input data policy module")
        moduleInstance = moduleFactory.getModule(inputDataPolicy, argumentsDict)
        if not moduleInstance["OK"]:
            return moduleInstance

        module = moduleInstance["Value"]
        result = module.execute()
        if not result["OK"]:
            self.log.warn("Input data resolution failed")
            return result

        return S_OK()

    #############################################################################
    def __checkFileCatalog(self, lfns, optReplicaInfo=None):
        """This function returns dictionaries containing all relevant parameters
        to allow data access from the relevant file catalogue.  Optionally, optimizer
        parameters can be supplied here but if these are not sufficient, the file catalogue
        is subsequently consulted.

        N.B. this will be considerably simplified when the DMS evolves to have a
        generic FC interface and a single call for all available information.
        """
        replicas = optReplicaInfo
        if not replicas:
            replicas = self.__getReplicaMetadata(lfns)
            if not replicas["OK"]:
                return replicas

        self.log.verbose(replicas)

        failedGUIDs = []
        for lfn, reps in replicas["Value"]["Successful"].items():  # can be an iterator
            if "GUID" not in reps:
                failedGUIDs.append(lfn)

        if failedGUIDs:
            self.log.info("The following file(s) were found not to have a GUID:\n%s" % ",\n".join(failedGUIDs))

        if failedGUIDs:
            return S_ERROR("File metadata is not available")
        return replicas

    #############################################################################
    def __getReplicaMetadata(self, lfns):
        """Wrapper function to consult catalog for all necessary file metadata
        and check the result.
        """
        start = time.time()
        # We are in a job, therefore interested in replicas for jobs
        repsResult = self.dm.getReplicasForJobs(lfns)
        timing = time.time() - start
        self.log.info("Replica Lookup Time: %.2f seconds " % (timing))
        if not repsResult["OK"]:
            self.log.warn(repsResult["Message"])
            return repsResult

        badLFNCount = 0
        badLFNs = []
        catalogResult = repsResult["Value"]

        for lfn, cause in catalogResult.get("Failed", {}).items():  # can be an iterator
            badLFNCount += 1
            badLFNs.append("LFN:%s Problem: %s" % (lfn, cause))

        for lfn, replicas in catalogResult.get("Successful", {}).items():  # can be an iterator
            if not replicas:
                badLFNCount += 1
                badLFNs.append("LFN:%s Problem: Null replica value" % (lfn))

        if badLFNCount:
            self.log.warn("Job Wrapper found %s problematic LFN(s) for job %s" % (badLFNCount, self.jobID))
            param = "\n".join(badLFNs)
            self.log.info(param)
            self.__setJobParam("MissingLFNs", param)
            return S_ERROR(JobMinorStatus.INPUT_NOT_AVAILABLE)

        # Must retrieve GUIDs from FC for files
        start = time.time()
        guidDict = self.fc.getFileMetadata(lfns)
        timing = time.time() - start
        self.log.info("GUID Lookup Time: %.2f seconds " % (timing))
        if not guidDict["OK"]:
            self.log.warn("Failed to retrieve GUIDs from file catalog")
            self.log.warn(guidDict["Message"])
            return guidDict

        failed = guidDict["Value"]["Failed"]
        if failed:
            self.log.warn("Could not retrieve GUIDs from catalog for the following files")
            self.log.warn(failed)
            return S_ERROR("Missing GUIDs")

        for lfn, reps in repsResult["Value"]["Successful"].items():  # can be an iterator
            guidDict["Value"]["Successful"][lfn].update(reps)

        catResult = guidDict
        return catResult

    #############################################################################
    def processJobOutputs(self):
        """Outputs for a job may be treated here."""

        # first iteration of this, no checking of wildcards or oversize sandbox files etc.
        outputSandbox = self.jobArgs.get("OutputSandbox", [])
        if isinstance(outputSandbox, six.string_types):
            outputSandbox = [outputSandbox]
        if outputSandbox:
            self.log.verbose("OutputSandbox files are: %s" % ", ".join(outputSandbox))
        outputData = self.jobArgs.get("OutputData", [])
        if outputData and isinstance(outputData, six.string_types):
            outputData = outputData.split(";")
        if outputData:
            self.log.verbose("OutputData files are: %s" % ", ".join(outputData))

        # First resolve any wildcards for output files and work out if any files are missing
        resolvedSandbox = self.__resolveOutputSandboxFiles(outputSandbox)
        if not resolvedSandbox["OK"]:
            self.log.warn("Output sandbox file resolution failed:")
            self.log.warn(resolvedSandbox["Message"])
            self.__report(status=JobStatus.FAILED, minorStatus=JobMinorStatus.RESOLVING_OUTPUT_SANDBOX)

        fileList = resolvedSandbox["Value"]["Files"]
        missingFiles = resolvedSandbox["Value"]["Missing"]
        if missingFiles:
            self.jobReport.setJobParameter("OutputSandboxMissingFiles", ", ".join(missingFiles), sendFlag=False)

        if "Owner" not in self.jobArgs:
            msg = "Job has no owner specified"
            self.log.warn(msg)
            return S_OK(msg)

        # Do not overwrite in case of Error
        if not self.failedFlag:
            self.__report(status=JobStatus.COMPLETING, minorStatus=JobMinorStatus.UPLOADING_OUTPUT_SANDBOX)

        uploadOutputDataInAnyCase = False

        if fileList and self.jobID:
            self.outputSandboxSize = getGlobbedTotalSize(fileList)
            self.log.info("Attempting to upload Sandbox with limit:", self.sandboxSizeLimit)
            sandboxClient = SandboxStoreClient()
            result_sbUpload = sandboxClient.uploadFilesAsSandboxForJob(
                fileList, self.jobID, "Output", self.sandboxSizeLimit
            )  # 1024*1024*10
            if not result_sbUpload["OK"]:
                self.log.error("Output sandbox upload failed with message", result_sbUpload["Message"])
                outputSandboxData = result_sbUpload.get("SandboxFileName")
                if outputSandboxData:

                    self.log.info("Attempting to upload %s as output data" % (outputSandboxData))
                    if self.failedFlag:
                        outputData = [outputSandboxData]
                        uploadOutputDataInAnyCase = True
                    else:
                        outputData.append(outputSandboxData)
                    self.jobReport.setJobParameter("OutputSandbox", "Sandbox uploaded to grid storage", sendFlag=False)
                    self.jobReport.setJobParameter(
                        "OutputSandboxLFN", self.__getLFNfromOutputFile(outputSandboxData)[0], sendFlag=False
                    )
                else:
                    self.log.info("Could not get SandboxFileName to attempt upload to Grid storage")
                    return S_ERROR(
                        "Output sandbox upload failed and no file name supplied for failover to Grid storage"
                    )
            else:
                # Do not overwrite in case of Error
                if not self.failedFlag:
                    self.__report(status=JobStatus.COMPLETING, minorStatus=JobMinorStatus.OUTPUT_SANDBOX_UPLOADED)
                self.log.info("Sandbox uploaded successfully")

        if (outputData and not self.failedFlag) or uploadOutputDataInAnyCase:
            # Do not upload outputdata if the job has failed.
            # The exception is when the outputData is what was the OutputSandbox, which should be uploaded in any case
            outputSE = self.jobArgs.get("OutputSE", self.defaultOutputSE)
            if isinstance(outputSE, six.string_types):
                outputSE = [outputSE]

            outputPath = self.jobArgs.get("OutputPath", self.defaultOutputPath)
            if not isinstance(outputPath, six.string_types):
                outputPath = self.defaultOutputPath

            if not outputSE and not self.defaultFailoverSE:
                return S_ERROR("No output SEs defined in VO configuration")

            result_transferODF = self.__transferOutputDataFiles(outputData, outputSE, outputPath)

            # now that we (tried to) transfer the output files,
            # including possibly oversized Output Sandboxes,
            # we delete the local output sandbox tarfile in case it's still there.
            if not result_sbUpload["OK"]:
                outputSandboxData = result_sbUpload.get("SandboxFileName")
                if outputSandboxData:
                    try:
                        os.unlink(outputSandboxData)
                    except OSError:
                        pass

            if not result_transferODF["OK"]:
                return result_transferODF

        return S_OK("Job outputs processed")

    #############################################################################
    def __resolveOutputSandboxFiles(self, outputSandbox):
        """Checks the output sandbox file list and resolves any specified wildcards.
        Also tars any specified directories.
        """
        missing = []
        okFiles = []
        for i in outputSandbox:
            self.log.verbose("Looking at OutputSandbox file/directory/wildcard: %s" % i)
            globList = glob.glob(i)
            for check in globList:
                if os.path.isfile(check):
                    self.log.verbose("Found locally existing OutputSandbox file: %s" % check)
                    okFiles.append(check)
                if os.path.isdir(check):
                    self.log.verbose("Found locally existing OutputSandbox directory: %s" % check)
                    cmd = ["tar", "cf", "%s.tar" % check, check]
                    result = systemCall(60, cmd)
                    if not result["OK"]:
                        self.log.error("Failed to create OutputSandbox tar", result["Message"])
                    elif result["Value"][0]:
                        self.log.error("Failed to create OutputSandbox tar", result["Value"][2])
                    if os.path.isfile("%s.tar" % (check)):
                        self.log.verbose("Appending %s.tar to OutputSandbox" % check)
                        okFiles.append("%s.tar" % (check))
                    else:
                        self.log.warn("Could not tar OutputSandbox directory: %s" % check)
                        missing.append(check)

        for i in outputSandbox:
            if i not in okFiles:
                if not "%s.tar" % i in okFiles:
                    if not re.search(r"\*", i):
                        if i not in missing:
                            missing.append(i)

        result = {"Missing": missing, "Files": okFiles}
        return S_OK(result)

    #############################################################################
    def __transferOutputDataFiles(self, outputData, outputSE, outputPath):
        """Performs the upload and registration in the File Catalog(s)"""
        self.log.verbose("Uploading output data files")
        self.__report(status=JobStatus.COMPLETING, minorStatus=JobMinorStatus.UPLOADING_OUTPUT_DATA)
        self.log.info("Output data files %s to be uploaded to %s SE" % (", ".join(outputData), outputSE))
        missing = []
        uploaded = []

        # Separate outputdata in the form of lfns and local files
        lfnList = []
        nonlfnList = []
        for out in outputData:
            if out.lower().find("lfn:") != -1:
                lfnList.append(out)
            else:
                nonlfnList.append(out)

        # Check whether list of outputData has a globbable pattern
        globbedOutputList = List.uniqueElements(getGlobbedFiles(nonlfnList))
        if globbedOutputList != nonlfnList and globbedOutputList:
            self.log.info(
                "Found a pattern in the output data file list, files to upload are:", ", ".join(globbedOutputList)
            )
            nonlfnList = globbedOutputList
        outputData = lfnList + nonlfnList

        pfnGUID = {}
        result = getGUID(outputData)
        if not result["OK"]:
            self.log.warn(
                "Failed to determine POOL GUID(s) for output file list (OK if not POOL files)", result["Message"]
            )
        else:
            pfnGUID = result["Value"]

        for outputFile in outputData:
            (lfn, localfile) = self.__getLFNfromOutputFile(outputFile, outputPath)
            if not os.path.exists(localfile):
                self.log.error("Missing specified output data file:", outputFile)
                continue

            # # file size
            localfileSize = getGlobbedTotalSize(localfile)

            self.outputDataSize += getGlobbedTotalSize(localfile)

            outputFilePath = os.path.join(os.getcwd(), localfile)

            # # file GUID
            fileGUID = pfnGUID[localfile] if localfile in pfnGUID else None
            if fileGUID:
                self.log.verbose("Found GUID for file from POOL XML catalogue %s" % localfile)

            # #  file checksum
            cksm = fileAdler(outputFilePath)

            fileMetaDict = {
                "Size": localfileSize,
                "LFN": lfn,
                "ChecksumType": "Adler32",
                "Checksum": cksm,
                "GUID": fileGUID,
            }

            outputSEList = self.__getSortedSEList(outputSE)
            upload = self.failoverTransfer.transferAndRegisterFile(
                fileName=localfile,
                localPath=outputFilePath,
                lfn=lfn,
                destinationSEList=outputSEList,
                fileMetaDict=fileMetaDict,
                fileCatalog=self.defaultCatalog,
                masterCatalogOnly=self.masterCatalogOnlyFlag,
                retryUpload=self.retryUpload,
            )
            if upload["OK"]:
                self.log.info(
                    '"%s" successfully uploaded to "%s" as "LFN:%s"' % (localfile, upload["Value"]["uploadedSE"], lfn)
                )
                uploaded.append(lfn)
                continue

            self.log.error(
                "Could not putAndRegister file",
                "%s with LFN %s to %s with GUID %s trying failover storage"
                % (localfile, lfn, ", ".join(outputSEList), fileGUID),
            )
            if not self.defaultFailoverSE:
                self.log.info(
                    "No failover SEs defined for JobWrapper,",
                    "cannot try to upload output file %s anywhere else." % outputFile,
                )
                missing.append(outputFile)
                continue

            failoverSEs = self.__getSortedSEList(self.defaultFailoverSE)
            targetSE = outputSEList[0]
            result = self.failoverTransfer.transferAndRegisterFileFailover(
                fileName=localfile,
                localPath=outputFilePath,
                lfn=lfn,
                targetSE=targetSE,
                failoverSEList=failoverSEs,
                fileMetaDict=fileMetaDict,
                fileCatalog=self.defaultCatalog,
                masterCatalogOnly=self.masterCatalogOnlyFlag,
            )
            if not result["OK"]:
                self.log.error("Completely failed to upload file to failover SEs", result["Message"])
                missing.append(outputFile)
            else:
                self.log.info("File %s successfully uploaded to failover storage element" % lfn)
                uploaded.append(lfn)

        # For files correctly uploaded must report LFNs to job parameters
        if uploaded:
            report = ", ".join(uploaded)
            # In case the VO payload has also uploaded data using the same parameter
            # name this should be checked prior to setting.
            result = JobMonitoringClient().getJobParameter(int(self.jobID), "UploadedOutputData")
            if result["OK"]:
                if "UploadedOutputData" in result["Value"]:
                    report += ", %s" % result["Value"]["UploadedOutputData"]

            self.jobReport.setJobParameter("UploadedOutputData", report, sendFlag=False)

        # TODO Notify the user of any output data / output sandboxes
        if missing:
            self.__setJobParam("OutputData", "MissingFiles: %s" % ", ".join(missing))
            self.__report(status=JobStatus.FAILED, minorStatus=JobMinorStatus.UPLOADING_OUTPUT_DATA)
            return S_ERROR("Failed to upload OutputData")

        self.__report(status=JobStatus.COMPLETING, minorStatus=JobMinorStatus.OUTPUT_DATA_UPLOADED)
        return S_OK("OutputData uploaded successfully")

    #############################################################################
    def __getSortedSEList(self, seList):
        """Randomize SE, putting first those that are Local/Close to the Site"""
        if not seList:
            return seList

        localSEs = []
        otherSEs = []
        siteSEs = []
        seMapping = getSEsForSite(DIRAC.siteName())

        if seMapping["OK"] and seMapping["Value"]:
            siteSEs = seMapping["Value"]

        for seName in seList:
            if seName in siteSEs:
                localSEs.append(seName)
            else:
                otherSEs.append(seName)

        return List.randomize(localSEs) + List.randomize(otherSEs)

    #############################################################################
    def __getLFNfromOutputFile(self, outputFile, outputPath=""):
        """Provides a generic convention for VO output data
        files if no path is specified.
        """

        if not re.search("^LFN:", outputFile):
            localfile = outputFile
            initial = self.owner[:1]
            vo = getVOForGroup(self.userGroup)
            if not vo:
                vo = "dirac"

            ops = Operations(vo=vo)
            user_prefix = ops.getValue("LFNUserPrefix", "user")
            basePath = "/" + vo + "/" + user_prefix + "/" + initial + "/" + self.owner
            if outputPath:
                # If output path is given, append it to the user path and put output files in this directory
                if outputPath.startswith("/"):
                    outputPath = outputPath[1:]
            else:
                # By default the output path is constructed from the job id
                subdir = str(int(self.jobID / 1000))
                outputPath = subdir + "/" + str(self.jobID)
            lfn = os.path.join(basePath, outputPath, os.path.basename(localfile))
        else:
            # if LFN is given, take it as it is
            localfile = os.path.basename(outputFile.replace("LFN:", ""))
            lfn = outputFile.replace("LFN:", "")

        return (lfn, localfile)

    #############################################################################
    def transferInputSandbox(self, inputSandbox):
        """Downloads the input sandbox for the job"""
        sandboxFiles = []
        registeredISB = []
        lfns = []
        self.__report(status=JobStatus.RUNNING, minorStatus=JobMinorStatus.DOWNLOADING_INPUT_SANDBOX)
        if not isinstance(inputSandbox, (list, tuple)):
            inputSandbox = [inputSandbox]
        for isb in inputSandbox:
            if isb.find("LFN:") == 0 or isb.find("lfn:") == 0:
                lfns.append(isb)
            else:
                if isb.find("SB:") == 0:
                    registeredISB.append(isb)
                else:
                    sandboxFiles.append(os.path.basename(isb))

        self.log.info("Downloading InputSandbox for job %s: %s" % (self.jobID, ", ".join(sandboxFiles)))
        if os.path.exists("%s/inputsandbox" % (self.root)):
            # This is a debugging tool, get the file from local storage to debug Job Wrapper
            sandboxFiles.append("jobDescription.xml")
            for inputFile in sandboxFiles:
                if os.path.exists("%s/inputsandbox/%s" % (self.root, inputFile)):
                    self.log.info("Getting InputSandbox file %s from local directory for testing" % (inputFile))
                    shutil.copy(self.root + "/inputsandbox/" + inputFile, inputFile)
            result = S_OK(sandboxFiles)
        else:
            if registeredISB:
                for isb in registeredISB:
                    self.log.info("Downloading Input SandBox %s" % isb)
                    result = SandboxStoreClient().downloadSandbox(isb)
                    if not result["OK"]:
                        self.__report(
                            status=JobStatus.RUNNING, minorStatus=JobMinorStatus.FAILED_DOWNLOADING_INPUT_SANDBOX
                        )
                        return S_ERROR("Cannot download Input sandbox %s: %s" % (isb, result["Message"]))
                    else:
                        self.inputSandboxSize += result["Value"]

        if lfns:
            self.log.info("Downloading Input SandBox LFNs, number of files to get", len(lfns))
            self.__report(status=JobStatus.RUNNING, minorStatus=JobMinorStatus.DOWNLOADING_INPUT_SANDBOX_LFN)
            lfns = [fname.replace("LFN:", "").replace("lfn:", "") for fname in lfns]
            download = self.dm.getFile(lfns)
            if not download["OK"]:
                self.log.warn(download)
                self.__report(status=JobStatus.RUNNING, minorStatus=JobMinorStatus.FAILED_DOWNLOADING_INPUT_SANDBOX_LFN)
                return S_ERROR(download["Message"])
            failed = download["Value"]["Failed"]
            if failed:
                self.log.warn("Could not download InputSandbox LFN(s)")
                self.log.warn(failed)
                return S_ERROR(str(failed))
            for lfn in lfns:
                if os.path.exists("%s/%s" % (self.root, os.path.basename(download["Value"]["Successful"][lfn]))):
                    sandboxFiles.append(os.path.basename(download["Value"]["Successful"][lfn]))

        userFiles = sandboxFiles + [os.path.basename(lfn) for lfn in lfns]
        for possibleTarFile in userFiles:
            if not os.path.exists(possibleTarFile):
                continue
            try:
                if os.path.isfile(possibleTarFile) and tarfile.is_tarfile(possibleTarFile):
                    self.log.info("Unpacking input sandbox file %s" % (possibleTarFile))
                    with tarfile.open(possibleTarFile, "r") as tarFile:
                        for member in tarFile.getmembers():
                            tarFile.extract(member, os.getcwd())
            except Exception as x:
                return S_ERROR("Could not untar %s with exception %s" % (possibleTarFile, str(x)))

        if userFiles:
            self.inputSandboxSize = getGlobbedTotalSize(userFiles)
            self.log.info(
                "Total size of input sandbox:",
                "%0.2f MiB (%s bytes)" % (self.inputSandboxSize / 1048576.0, self.inputSandboxSize),
            )

        return S_OK("InputSandbox downloaded")

    #############################################################################
    def finalize(self):
        """Perform any final actions to clean up after job execution."""
        self.log.info("Running JobWrapper finalization")

        # find if there are pending failover requests
        requests = self.__getRequestFiles()
        outputDataRequest = self.failoverTransfer.getRequest()

        requestFlag = len(requests) > 0 or not outputDataRequest.isEmpty()

        finalStatus = ""
        finalMinorStatus = ""
        prString = "Job finished "
        if self.failedFlag:
            prString += "with errors"
            finalStatus = JobStatus.FAILED
        else:
            prString += "successfully"
            if requestFlag:
                finalStatus = JobStatus.COMPLETED
            else:
                finalStatus = JobStatus.DONE
                finalMinorStatus = JobMinorStatus.EXEC_COMPLETE

        # send the failover request if any (also for failed jobs)
        res = self.sendFailoverRequest()
        if not res["OK"]:  # This means that the request could not be set (this should "almost never" happen)
            finalStatus = JobStatus.FAILED
            finalMinorStatus = JobMinorStatus.FAILED_SENDING_REQUESTS
            self.failedFlag = True
        elif res["Value"]:
            # A request was sent, change the minor status
            finalMinorStatus = JobMinorStatus.PENDING_REQUESTS
            requestFlag = True

        # Set the final status of the job
        self.log.info(prString, "with%s pending requests" % ("" if requestFlag else " no"))
        self.log.info("Final job status", "%s ; %s" % (finalStatus, finalMinorStatus))
        self.__report(status=finalStatus, minorStatus=finalMinorStatus, sendFlag=True)

        # Sending the last accounting report
        if not self.jobID:
            self.log.debug("No accounting to be sent since running locally")
        else:
            # Final status and minorStatus are already set
            result = self.sendJobAccounting()
            if not result["OK"]:
                # This should be really rare, as the accounting report also sends failover requests
                self.log.warn(
                    "JobAccountingFailure", "Could not send job accounting with result: \n%s" % result["Message"]
                )
                self.log.warn(
                    "JobAccountingFailure", "The job won't fail, but the accounting for this job won't be sent"
                )

        self.__cleanUp()
        return 1 if self.failedFlag else 0

    #############################################################################
    def sendJobAccounting(self, status="", minorStatus=""):
        """Send WMS accounting data."""
        if self.jobAccountingSent:
            return S_OK()
        if status:
            self.wmsMajorStatus = status
        if minorStatus:
            self.wmsMinorStatus = minorStatus

        self.accountingReport.setEndTime()
        # CPUTime and ExecTime
        if "CPU" not in EXECUTION_RESULT:
            # If the payload has not started execution (error with input data, SW, SB,...)
            # Execution result is not filled use self.initialTiming
            self.log.info("EXECUTION_RESULT[CPU] missing in sendJobAccounting")
            finalStat = os.times()
            EXECUTION_RESULT["CPU"] = []
            for i, _ in enumerate(finalStat):
                EXECUTION_RESULT["CPU"].append(finalStat[i] - self.initialTiming[i])

        cpuString = " ".join(["%.2f" % x for x in EXECUTION_RESULT["CPU"]])
        self.log.info("EXECUTION_RESULT[CPU] in sendJobAccounting", cpuString)

        utime, stime, cutime, cstime, elapsed = EXECUTION_RESULT["CPU"]
        cpuTime = utime + stime + cutime + cstime
        execTime = elapsed
        diskSpaceConsumed = getGlobbedTotalSize(os.path.join(self.root, str(self.jobID)))
        # Fill the data
        acData = {
            "User": self.owner,
            "UserGroup": self.userGroup,
            "JobGroup": self.jobGroup,
            "JobType": self.jobType,
            "JobClass": self.jobClass,
            "ProcessingType": self.processingType,
            "FinalMajorStatus": self.wmsMajorStatus,
            "FinalMinorStatus": self.wmsMinorStatus,
            "CPUTime": cpuTime,
            # Based on the factor to convert raw CPU to Normalized units (based on the CPU Model)
            "NormCPUTime": cpuTime * self.cpuNormalizationFactor,
            "ExecTime": execTime * self.numberOfProcessors,
            "InputDataSize": self.inputDataSize,
            "OutputDataSize": self.outputDataSize,
            "InputDataFiles": self.inputDataFiles,
            "OutputDataFiles": self.outputDataFiles,
            "DiskSpace": diskSpaceConsumed,
            "InputSandBoxSize": self.inputSandboxSize,
            "OutputSandBoxSize": self.outputSandboxSize,
            "ProcessedEvents": self.processedEvents,
        }
        self.log.verbose("Accounting Report is:")
        self.log.verbose(acData)
        self.accountingReport.setValuesFromDict(acData)
        result = self.accountingReport.commit()
        # Even if it fails a failover request will be created
        self.jobAccountingSent = True
        return result

    #############################################################################
    def sendFailoverRequest(self):
        """Create and send a combined job failover request if any"""
        request = Request()
        # Forbid the request to be executed within the requested delay
        request.delayNextExecution(self.failoverRequestDelay)

        requestName = "job_%s" % self.jobID
        if "JobName" in self.jobArgs:
            # To make the request names more appealing for users
            jobName = self.jobArgs["JobName"]
            if isinstance(jobName, six.string_types) and jobName:
                jobName = jobName.replace(" ", "").replace("(", "").replace(")", "").replace('"', "")
                jobName = jobName.replace(".", "").replace("{", "").replace("}", "").replace(":", "")
                requestName = "%s_%s" % (jobName, requestName)

        request.RequestName = requestName.replace('"', "")
        request.JobID = self.jobID
        request.SourceComponent = "Job_%s" % self.jobID

        # JobReport part first
        result = self.jobReport.generateForwardDISET()
        if result["OK"]:
            if isinstance(result["Value"], Operation):
                self.log.info("Adding a job state update DISET operation to the request")
                request.addOperation(result["Value"])
        else:
            self.log.warn("JobReportFailure", "Could not generate a forwardDISET operation: %s" % result["Message"])
            self.log.warn("JobReportFailure", "The job won't fail, but the jobLogging info might be incomplete")

        # Failover transfer requests
        for storedOperation in self.failoverTransfer.request:
            request.addOperation(storedOperation)

        # Any other requests in the current directory
        rfiles = self.__getRequestFiles()
        for rfname in rfiles:
            with open(rfname, "r") as rFile:
                requestStored = Request(json.load(rFile))
            for storedOperation in requestStored:
                request.addOperation(storedOperation)

        if len(request):
            # The request is ready, send it now
            isValid = RequestValidator().validate(request)
            if not isValid["OK"]:
                self.log.error("Failover request is not valid", isValid["Message"])
                self.log.error("Job will fail, first trying to print out the content of the request")
                reqToJSON = request.toJSON()
                if reqToJSON["OK"]:
                    print(str(reqToJSON["Value"]))
                else:
                    self.log.error("Something went wrong creating the JSON from request", reqToJSON["Message"])
                return S_ERROR()
            else:
                # We try several times to put the request before failing the job:
                # it is very important that requests go through,
                # or the job will be in an unclear status
                # (workflow ok, but, e.g., the output files won't be registered).
                # It's a poor man solution, but I don't see fancy alternatives
                for counter in range(10):
                    requestClient = ReqClient()
                    result = requestClient.putRequest(request)
                    if result["OK"]:
                        resDigest = request.getDigest()
                        digest = resDigest["Value"]
                        self.jobReport.setJobParameter("PendingRequest", digest)
                        return S_OK(request)
                    else:
                        self.log.error(
                            "Failed to set failover request", "%d: %s. Re-trying..." % (counter, result["Message"])
                        )
                        del requestClient
                        time.sleep(counter ** 3)

                if not result["OK"]:
                    return result

        return S_OK()

    #############################################################################
    def __getRequestFiles(self):
        """Simple wrapper to return the list of request files."""
        return glob.glob("*_request.json")

    #############################################################################
    def __cleanUp(self):
        """Cleans up after job processing. Can be switched off via environment
        variable DO_NOT_DO_JOB_CLEANUP or by JobWrapper configuration option.
        """
        # Environment variable is a feature for DIRAC (helps local debugging).
        if "DO_NOT_DO_JOB_CLEANUP" in os.environ or not self.cleanUpFlag:
            cleanUp = False
        else:
            cleanUp = True

        os.chdir(self.root)
        if cleanUp:
            self.log.verbose("Cleaning up job working directory")
            if os.path.exists(str(self.jobID)):
                shutil.rmtree(str(self.jobID))

    #############################################################################
    def __report(self, status="", minorStatus="", sendFlag=False):
        """Wraps around setJobStatus of jobReport object"""
        self.log.verbose("setJobStatus", "(%s,%s,%s,%s)" % (self.jobID, status, minorStatus, "JobWrapper"))

        if status:
            self.wmsMajorStatus = status
        if minorStatus:
            self.wmsMinorStatus = minorStatus
        jobStatus = self.jobReport.setJobStatus(status=status, minorStatus=minorStatus, sendFlag=sendFlag)
        if not jobStatus["OK"]:
            self.log.warn("Failed setting job status", jobStatus["Message"])

        return jobStatus

    #############################################################################
    def __setJobParam(self, name, value, sendFlag=False):
        """Wraps around setJobParameter of JobReport client"""
        jobParam = self.jobReport.setJobParameter(str(name), str(value), sendFlag)
        if not jobParam["OK"]:
            self.log.warn("Failed setting job parameter", jobParam["Message"])
        if self.jobID:
            self.log.verbose("setJobParameter", "(%s,%s,%s)" % (self.jobID, name, value))

        return jobParam

    #############################################################################
    def __setJobParamList(self, value, sendFlag=False):
        """Wraps around setJobParameters of JobReport client"""
        jobParam = self.jobReport.setJobParameters(value, sendFlag)
        if not jobParam["OK"]:
            self.log.warn(jobParam["Message"])
        if self.jobID:
            self.log.verbose("setJobParameters(%s,%s)" % (self.jobID, value))

        return jobParam


###############################################################################
###############################################################################


class ExecutionThread(threading.Thread):

    #############################################################################
    def __init__(self, spObject, cmd, maxPeekLines, stdoutFile, stderrFile, exeEnv):
        threading.Thread.__init__(self)
        self.cmd = cmd
        self.spObject = spObject
        self.outputLines = []
        self.maxPeekLines = maxPeekLines
        self.stdout = stdoutFile
        self.stderr = stderrFile
        self.exeEnv = exeEnv

    #############################################################################
    def run(self):
        """Method representing the thread activity.
        This one overrides the ~threading.Thread `run` method
        """
        log = gLogger.getSubLogger("ExecutionThread")

        # FIXME: why local instances of object variables are created?
        cmd = self.cmd
        spObject = self.spObject
        start = time.time()
        initialStat = os.times()
        log.verbose("Cmd called", cmd)
        output = spObject.systemCall(cmd, env=self.exeEnv, callbackFunction=self.sendOutput, shell=True)
        log.verbose("Output of system call within execution thread: %s" % output)
        EXECUTION_RESULT["Thread"] = output
        timing = time.time() - start
        EXECUTION_RESULT["Timing"] = timing
        finalStat = os.times()
        EXECUTION_RESULT["CPU"] = []
        for i, _ in enumerate(finalStat):
            EXECUTION_RESULT["CPU"].append(finalStat[i] - initialStat[i])
        cpuString = " ".join(["%.2f" % x for x in EXECUTION_RESULT["CPU"]])
        log.info("EXECUTION_RESULT[CPU] after Execution of spObject.systemCall", cpuString)
        log.info("EXECUTION_RESULT[Thread] after Execution of spObject.systemCall", str(EXECUTION_RESULT["Thread"]))

    #############################################################################
    def getCurrentPID(self):
        return self.spObject.getChildPID()

    #############################################################################
    def sendOutput(self, stdid, line):
        if stdid == 0 and self.stdout:
            with open(self.stdout, "a+") as outputFile:
                print(line, file=outputFile)
        elif stdid == 1 and self.stderr:
            with open(self.stderr, "a+") as errorFile:
                print(line, file=errorFile)
        self.outputLines.append(line)
        size = len(self.outputLines)
        if size > self.maxPeekLines:
            # reduce max size of output peeking
            self.outputLines.pop(0)

    #############################################################################
    def getOutput(self, lines=0):
        if self.outputLines:
            # restrict to smaller number of lines for regular
            # peeking by the watchdog
            # FIXME: this is multithread, thus single line would be better
            if lines:
                size = len(self.outputLines)
                cut = size - lines
                self.outputLines = self.outputLines[cut:]
            return S_OK(self.outputLines)
        return S_ERROR("No Job output found")


def rescheduleFailedJob(jobID, minorStatus, jobReport=None):
    """Function for rescheduling a jobID, setting a minorStatus"""

    rescheduleResult = JobStatus.RESCHEDULED

    try:

        gLogger.warn("Failure during %s" % (minorStatus))

        # Setting a job parameter does not help since the job will be rescheduled,
        # instead set the status with the cause and then another status showing the
        # reschedule operation.

        if not jobReport:
            gLogger.info("Creating a new JobReport Object")
            jobReport = JobReport(int(jobID), "JobWrapper")

        jobReport.setApplicationStatus("Failed %s " % minorStatus, sendFlag=False)
        jobReport.setJobStatus(status=JobStatus.RESCHEDULED, minorStatus=minorStatus, sendFlag=False)

        # We must send Job States and Parameters before it gets reschedule
        jobReport.sendStoredStatusInfo()
        jobReport.sendStoredJobParameters()

        gLogger.info("Job will be rescheduled after exception during execution of the JobWrapper")

        jobManager = JobManagerClient()
        result = jobManager.rescheduleJob(int(jobID))
        if not result["OK"]:
            gLogger.warn(result["Message"])
            if "Maximum number of reschedulings is reached" in result["Message"]:
                rescheduleResult = JobStatus.FAILED

        return rescheduleResult
    except Exception:
        gLogger.exception("JobWrapperTemplate failed to reschedule Job")
        return JobStatus.FAILED



"""HTCondorCE Computing Element

Allows direct submission to HTCondorCE Computing Elements with a SiteDirector Agent

**Configuration Parameters**

Configuration for the HTCondorCE submission can be done via the configuration system. See the page about
configuring :ref:`resourcesComputing` for where the options can be placed.

WorkingDirectory:
   Location to store the pilot and condor log files locally. It should exist on the server and be accessible (both
   readable and writeable).  Also temporary files like condor submit files are kept here. This option is only read
   from the global Resources/Computing/HTCondorCE location.

DaysToKeepRemoteLogs:
   How long to keep the log files on the remote schedd until they are removed

DaysToKeepLogs:
   How long to keep the log files locally until they are removed

ExtraSubmitString:
   Additional options for the condor submit file, separate options with '\\n', for example::

     request_cpus = 8 \\n periodic_remove = ...

   CERN proposes additional features to the standard HTCondor implementation. Among these features, one can find
   an option to limit the allocation runtime (`+MaxRuntime`), that does not exist in the standard HTCondor version:
   no explicit way to define a runtime limit (`maxCPUTime` would act as the limit). On CERN-HTCondor CEs, one can use
   CERN-specific features via the `ExtraSubmitString` configuration parameter.

UseLocalSchedd:
   If False, directly submit to a remote condor schedule daemon,
   then one does not need to run condor daemons on the submit machine.
   If True requires the condor grid middleware (condor_submit, condor_history, condor_q, condor_rm)

**Proxy renewal or lifetime**

When not using a local condor_schedd, add ``delegate_job_GSI_credentials_lifetime = 0`` to the ``ExtraSubmitString``.

When using a local condor_schedd look at the HTCondor documenation for enabling the proxy refresh.

**Code Documentation**
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Note: if you read this documentation in the source code and not via the sphinx
# created documentation, there should only be one slash when setting the option,
# but "\n" gets rendered as a linebreak in sphinx

import six
import os
import tempfile

# TODO: This should be modernised to use subprocess(32)
try:
    import commands
except ImportError:
    # Python 3's subprocess module contains a compatibility layer
    import subprocess as commands
import datetime
import errno
import threading
import json

from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.Resources.Computing.ComputingElement import ComputingElement
from DIRAC.Core.Utilities.Grid import executeGridCommand
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.WorkloadManagementSystem.Client import PilotStatus
from DIRAC.WorkloadManagementSystem.Client.PilotManagerClient import PilotManagerClient
from DIRAC.Core.Utilities.File import makeGuid
from DIRAC.Core.Utilities.Subprocess import Subprocess

from DIRAC.Resources.Computing.BatchSystems.Condor import parseCondorStatus, treatCondorHistory

__RCSID__ = "$Id$"

CE_NAME = "HTCondorCE"
MANDATORY_PARAMETERS = ["Queue"]
DEFAULT_WORKINGDIRECTORY = "/opt/dirac/pro/runit/WorkloadManagement/SiteDirectorHT"
DEFAULT_DAYSTOKEEPREMOTELOGS = 1
DEFAULT_DAYSTOKEEPLOGS = 15


def logDir(ceName, stamp):
    """Return path to log and output files for pilot.

    :param str ceName: Name of the CE
    :param str stamp: pilot stamp from/for jobRef
    """
    return os.path.join(ceName, stamp[0], stamp[1:3])


def condorIDAndPathToResultFromJobRef(jobRef):
    """Extract tuple of jobURL and jobID from the jobRef string.
    The condorID as well as the path leading to the job results are also extracted from the jobID.

    :param str jobRef: PilotJobReference of the following form: ``htcondorce://<ceName>/<condorID>:::<pilotStamp>``

    :return: tuple composed of the jobURL, the path to the job results and the condorID of the given jobRef
    """
    splits = jobRef.split(":::")
    jobURL = splits[0]
    stamp = splits[1] if len(splits) > 1 else ""
    _, _, ceName, condorID = jobURL.split("/")

    # Reconstruct the path leading to the result (log, output)
    # Construction of the path can be found in submitJob()
    pathToResult = logDir(ceName, stamp) if len(stamp) >= 3 else ""

    return jobURL, pathToResult, condorID


def findFile(workingDir, fileName, pathToResult):
    """Find a file in a file system.

    :param str workingDir: the name of the directory containing the given file to search for
    :param str fileName: the name of the file to find
    :param str pathToResult: the path to follow from workingDir to find the file

    :return: path leading to the file
    """
    path = os.path.join(workingDir, pathToResult, fileName)
    if os.path.exists(path):
        return S_OK(path)
    return S_ERROR(errno.ENOENT, "Could not find %s" % path)


class HTCondorCEComputingElement(ComputingElement):
    """HTCondorCE computing element class
    implementing the functions jobSubmit, getJobOutput
    """

    # static variables to ensure single cleanup every minute
    _lastCleanupTime = datetime.datetime.utcnow()
    _cleanupLock = threading.Lock()

    #############################################################################
    def __init__(self, ceUniqueID):
        """Standard constructor."""
        super(HTCondorCEComputingElement, self).__init__(ceUniqueID)

        self.ceType = CE_NAME
        self.submittedJobs = 0
        self.mandatoryParameters = MANDATORY_PARAMETERS
        self.pilotProxy = ""
        self.queue = ""
        self.outputURL = "gsiftp://localhost"
        self.gridEnv = ""
        self.proxyRenewal = 0
        self.daysToKeepLogs = DEFAULT_DAYSTOKEEPLOGS
        self.daysToKeepRemoteLogs = DEFAULT_DAYSTOKEEPREMOTELOGS
        self.extraSubmitString = ""
        # see note on getCondorLogFile, why we can only use the global setting
        self.workingDirectory = gConfig.getValue(
            "Resources/Computing/HTCondorCE/WorkingDirectory", DEFAULT_WORKINGDIRECTORY
        )
        self.useLocalSchedd = True
        self.remoteScheddOptions = ""

    #############################################################################
    def __writeSub(self, executable, nJobs, location, processors):
        """Create the Sub File for submission.

        :param str executable: name of the script to execute
        :param int nJobs: number of desired jobs
        :param str location: directory that should contain the result of the jobs
        :param int processors: number of CPU cores to allocate
        """

        self.log.debug("Working directory: %s " % self.workingDirectory)
        mkDir(os.path.join(self.workingDirectory, location))

        self.log.debug("InitialDir: %s" % os.path.join(self.workingDirectory, location))

        self.log.debug("ExtraSubmitString:\n### \n %s \n###" % self.extraSubmitString)

        fd, name = tempfile.mkstemp(suffix=".sub", prefix="HTCondorCE_", dir=self.workingDirectory)
        subFile = os.fdopen(fd, "w")

        executable = os.path.join(self.workingDirectory, executable)

        # This is used to remove outputs from the remote schedd
        # Used in case a local schedd is not used
        periodicRemove = "periodic_remove = "
        periodicRemove += "(JobStatus == 4) && "
        periodicRemove += "(time() - EnteredCurrentStatus) > (%s * 24 * 3600)" % self.daysToKeepRemoteLogs

        localScheddOptions = (
            """
ShouldTransferFiles = YES
WhenToTransferOutput = ON_EXIT_OR_EVICT
"""
            if self.useLocalSchedd
            else periodicRemove
        )

        targetUniverse = "grid" if self.useLocalSchedd else "vanilla"

        sub = """
executable = %(executable)s
universe = %(targetUniverse)s
use_x509userproxy = true
output = $(Cluster).$(Process).out
error = $(Cluster).$(Process).err
log = $(Cluster).$(Process).log
environment = "HTCONDOR_JOBID=$(Cluster).$(Process)"
initialdir = %(initialDir)s
grid_resource = condor %(ceName)s %(ceName)s:9619
transfer_output_files = ""
request_cpus = %(processors)s
%(localScheddOptions)s

kill_sig=SIGTERM

%(extraString)s

Queue %(nJobs)s

""" % dict(
            executable=executable,
            nJobs=nJobs,
            processors=processors,
            ceName=self.ceName,
            extraString=self.extraSubmitString,
            initialDir=os.path.join(self.workingDirectory, location),
            localScheddOptions=localScheddOptions,
            targetUniverse=targetUniverse,
        )
        subFile.write(sub)
        subFile.close()
        return name

    def _reset(self):
        self.queue = self.ceParameters["Queue"]
        self.outputURL = self.ceParameters.get("OutputURL", "gsiftp://localhost")
        self.gridEnv = self.ceParameters.get("GridEnv")
        self.daysToKeepLogs = self.ceParameters.get("DaysToKeepLogs", DEFAULT_DAYSTOKEEPLOGS)
        self.extraSubmitString = str(self.ceParameters.get("ExtraSubmitString", "").encode().decode("unicode_escape"))
        self.daysToKeepRemoteLogs = self.ceParameters.get("DaysToKeepRemoteLogs", DEFAULT_DAYSTOKEEPREMOTELOGS)
        self.useLocalSchedd = self.ceParameters.get("UseLocalSchedd", self.useLocalSchedd)
        if isinstance(self.useLocalSchedd, six.string_types):
            if self.useLocalSchedd == "False":
                self.useLocalSchedd = False

        self.remoteScheddOptions = "" if self.useLocalSchedd else "-pool %s:9619 -name %s " % (self.ceName, self.ceName)

        self.log.debug("Using local schedd: %r " % self.useLocalSchedd)
        self.log.debug("Remote scheduler option: '%s' " % self.remoteScheddOptions)
        return S_OK()

    #############################################################################
    def submitJob(self, executableFile, proxy, numberOfJobs=1):
        """Method to submit job"""

        self.log.verbose("Executable file path: %s" % executableFile)
        if not os.access(executableFile, 5):
            os.chmod(executableFile, 0o755)

        # The submitted pilots are going to have a common part of the stamp to construct a path to retrieve results
        # Then they also have an individual part to make them unique
        jobStamps = []
        commonJobStampPart = makeGuid()[:3]
        for _i in range(numberOfJobs):
            jobStamp = commonJobStampPart + makeGuid()[:5]
            jobStamps.append(jobStamp)

        # We randomize the location of the pilot output and log, because there are just too many of them
        location = logDir(self.ceName, commonJobStampPart)
        nProcessors = self.ceParameters.get("NumberOfProcessors", 1)
        subName = self.__writeSub(executableFile, numberOfJobs, location, nProcessors)

        cmd = ["condor_submit", "-terse", subName]
        # the options for submit to remote are different than the other remoteScheddOptions
        # -remote: submit to a remote condor_schedd and spool all the required inputs
        scheddOptions = [] if self.useLocalSchedd else ["-pool", "%s:9619" % self.ceName, "-remote", self.ceName]
        for op in scheddOptions:
            cmd.insert(-1, op)

        result = executeGridCommand(self.proxy, cmd, self.gridEnv)
        self.log.verbose(result)
        os.remove(subName)
        if not result["OK"]:
            self.log.error("Failed to submit jobs to htcondor", result["Message"])
            return result

        if result["Value"][0]:
            # We have got a non-zero status code
            errorString = result["Value"][2] if result["Value"][2] else result["Value"][1]
            return S_ERROR("Pilot submission failed with error: %s " % errorString.strip())

        pilotJobReferences = self.__getPilotReferences(result["Value"][1].strip())
        if not pilotJobReferences["OK"]:
            return pilotJobReferences
        pilotJobReferences = pilotJobReferences["Value"]

        self.log.verbose("JobStamps: %s " % jobStamps)
        self.log.verbose("pilotRefs: %s " % pilotJobReferences)

        result = S_OK(pilotJobReferences)
        result["PilotStampDict"] = dict(zip(pilotJobReferences, jobStamps))
        if self.useLocalSchedd:
            # Executable is transferred afterward
            # Inform the caller that Condor cannot delete it before the end of the execution
            result["ExecutableToKeep"] = executableFile

        self.log.verbose("Result for submission: %s " % result)
        return result

    def killJob(self, jobIDList):
        """Kill the specified jobs"""
        if not jobIDList:
            return S_OK()
        if isinstance(jobIDList, six.string_types):
            jobIDList = [jobIDList]

        self.log.verbose("KillJob jobIDList: %s" % jobIDList)

        for jobRef in jobIDList:
            job, _, jobID = condorIDAndPathToResultFromJobRef(jobRef)
            self.log.verbose("Killing pilot %s " % job)
            status, stdout = commands.getstatusoutput("condor_rm %s %s" % (self.remoteScheddOptions, jobID))
            if status != 0:
                return S_ERROR("Failed to kill pilot %s: %s" % (job, stdout))

        return S_OK()

    #############################################################################
    def getCEStatus(self):
        """Method to return information on running and pending jobs.

        Warning: information currently returned depends on the PilotManager and not HTCondor.
        Results might be wrong if pilots or jobs are submitted manually via the CE.
        """
        result = S_OK()
        result["SubmittedJobs"] = 0
        result["RunningJobs"] = 0
        result["WaitingJobs"] = 0

        # getWaitingPilots
        condDict = {"DestinationSite": self.ceName, "Status": PilotStatus.PILOT_WAITING_STATES}
        res = PilotManagerClient().countPilots(condDict)
        if res["OK"]:
            result["WaitingJobs"] = int(res["Value"])
        else:
            self.log.warn("Failure getting pilot count for %s: %s " % (self.ceName, res["Message"]))

        # getRunningPilots
        condDict = {"DestinationSite": self.ceName, "Status": PilotStatus.RUNNING}
        res = PilotManagerClient().countPilots(condDict)
        if res["OK"]:
            result["RunningJobs"] = int(res["Value"])
        else:
            self.log.warn("Failure getting pilot count for %s: %s " % (self.ceName, res["Message"]))

        return result

    def getJobStatus(self, jobIDList):
        """Get the status information for the given list of jobs"""
        # If we use a local schedd, then we have to cleanup executables regularly
        if self.useLocalSchedd:
            self.__cleanup()

        self.log.verbose("Job ID List for status: %s " % jobIDList)
        if isinstance(jobIDList, six.string_types):
            jobIDList = [jobIDList]

        resultDict = {}
        condorIDs = {}
        # Get all condorIDs so we can just call condor_q and condor_history once
        for jobRef in jobIDList:
            job, _, jobID = condorIDAndPathToResultFromJobRef(jobRef)
            condorIDs[job] = jobID

        qList = []
        for _condorIDs in breakListIntoChunks(condorIDs.values(), 100):

            # This will return a list of 1245.75 3
            status, stdout_q = commands.getstatusoutput(
                "condor_q %s %s -af:j JobStatus " % (self.remoteScheddOptions, " ".join(_condorIDs))
            )
            if status != 0:
                return S_ERROR(stdout_q)
            _qList = stdout_q.strip().split("\n")
            qList.extend(_qList)

            # FIXME: condor_history does only support j for autoformat from 8.5.3,
            # format adds whitespace for each field This will return a list of 1245 75 3
            # needs to cocatenate the first two with a dot
            condorHistCall = "condor_history %s %s -af ClusterId ProcId JobStatus" % (
                self.remoteScheddOptions,
                " ".join(_condorIDs),
            )

            treatCondorHistory(condorHistCall, qList)

        for job, jobID in condorIDs.items():

            pilotStatus = parseCondorStatus(qList, jobID)
            if pilotStatus == "HELD":
                # make sure the pilot stays dead and gets taken out of the condor_q
                _rmStat, _rmOut = commands.getstatusoutput("condor_rm %s %s " % (self.remoteScheddOptions, jobID))
                # self.log.debug( "condor job killed: job %s, stat %s, message %s " % ( jobID, rmStat, rmOut ) )
                pilotStatus = PilotStatus.ABORTED

            resultDict[job] = pilotStatus

        self.log.verbose("Pilot Statuses: %s " % resultDict)
        return S_OK(resultDict)

    def getJobLog(self, jobID):
        """Get pilot job logging info from HTCondor

        :param str jobID: pilot job identifier
        :return: string representing the logging info of a given pilot job
        """
        self.log.verbose("Getting logging info for jobID", jobID)
        result = self.__getJobOutput(jobID, ["logging"])
        if not result["OK"]:
            return result

        return S_OK(result["Value"]["logging"])

    def getJobOutput(self, jobID):
        """Get job outputs: output and error files from HTCondor

        :param str jobID: job identifier
        :return: tuple of strings containing output and error contents
        """
        self.log.verbose("Getting job output for jobID: %s " % jobID)
        result = self.__getJobOutput(jobID, ["output", "error"])
        if not result["OK"]:
            return result

        return S_OK((result["Value"]["output"], result["Value"]["error"]))

    def __getJobOutput(self, jobID, outTypes):
        """Get job outputs: output, error and logging files from HTCondor

        :param str jobID: job identifier
        :param list outTypes: output types targeted (output, error and/or logging)
        """
        _job, pathToResult, condorID = condorIDAndPathToResultFromJobRef(jobID)
        iwd = os.path.join(self.workingDirectory, pathToResult)

        try:
            mkDir(iwd)
        except OSError as e:
            errorMessage = "Failed to create the pilot output directory"
            self.log.exception(errorMessage, iwd)
            return S_ERROR(e.errno, "%s (%s)" % (errorMessage, iwd))

        if not self.useLocalSchedd:
            cmd = ["condor_transfer_data", "-pool", "%s:9619" % self.ceName, "-name", self.ceName, condorID]
            result = executeGridCommand(self.proxy, cmd, self.gridEnv)
            self.log.verbose(result)

            # Getting 'logging' without 'error' and 'output' is possible but will generate command errors
            # We do not check the command errors if we only want 'logging'
            if "error" in outTypes or "output" in outTypes:
                errorMessage = "Failed to get job output from htcondor"
                if not result["OK"]:
                    self.log.error(errorMessage, result["Message"])
                    return result
                # Even if result is OK, the actual exit code of cmd can still be an error
                if result["OK"] and result["Value"][0] != 0:
                    outMessage = result["Value"][1].strip()
                    errMessage = result["Value"][2].strip()
                    varMessage = outMessage + " " + errMessage
                    self.log.error(errorMessage, varMessage)
                    return S_ERROR("%s: %s" % (errorMessage, varMessage))

        outputsSuffix = {"output": "out", "error": "err", "logging": "log"}
        outputs = {}
        for output, suffix in outputsSuffix.items():
            resOut = findFile(self.workingDirectory, "%s.%s" % (condorID, suffix), pathToResult)
            if not resOut["OK"]:
                # Return an error if the output type was targeted, else we continue
                if output in outTypes:
                    self.log.error("Failed to find", "%s for condor job %s" % (output, jobID))
                    return resOut
                continue
            outputfilename = resOut["Value"]

            try:
                # If the output type is not targeted, then we directly remove the file
                if output in outTypes:
                    with open(outputfilename) as outputfile:
                        outputs[output] = outputfile.read()
                # If a local schedd is used, we cannot retrieve the outputs again if we delete them
                if not self.useLocalSchedd:
                    os.remove(outputfilename)
            except IOError as e:
                self.log.error("Failed to open", "%s file: %s" % (output, str(e)))
                return S_ERROR("Failed to get pilot %s" % output)

        return S_OK(outputs)

    def __getPilotReferences(self, jobString):
        """Get the references from the condor_submit output.
        Cluster ids look like " 107.0 - 107.0 " or " 107.0 - 107.4 "

        :param str jobString: the output of condor_submit

        :return: job references such as htcondorce://<CE name>/<path to result>-<clusterID>.<i>
        """
        self.log.verbose("getPilotReferences: %s" % jobString)
        clusterIDs = jobString.split("-")
        if len(clusterIDs) != 2:
            return S_ERROR("Something wrong with the condor_submit output: %s" % jobString)
        clusterIDs = [clu.strip() for clu in clusterIDs]
        self.log.verbose("Cluster IDs parsed: %s " % clusterIDs)
        try:
            clusterID = clusterIDs[0].split(".")[0]
            numJobs = clusterIDs[1].split(".")[1]
        except IndexError:
            return S_ERROR("Something wrong with the condor_submit output: %s" % jobString)
        cePrefix = "htcondorce://%s/" % self.ceName
        jobReferences = ["%s%s.%s" % (cePrefix, clusterID, i) for i in range(int(numJobs) + 1)]
        return S_OK(jobReferences)

    def __cleanup(self):
        """Clean the working directory of old jobs"""
        if not HTCondorCEComputingElement._cleanupLock.acquire(False):
            return

        now = datetime.datetime.utcnow()
        if (now - HTCondorCEComputingElement._lastCleanupTime).total_seconds() < 60:
            HTCondorCEComputingElement._cleanupLock.release()
            return

        HTCondorCEComputingElement._lastCleanupTime = now

        self.log.debug("Cleaning working directory: %s" % self.workingDirectory)

        # remove all files older than 120 minutes starting with DIRAC_ Condor will
        # push files on submission, but it takes at least a few seconds until this
        # happens so we can't directly unlink after condor_submit
        status, stdout = commands.getstatusoutput(
            'find -O3 %s -maxdepth 1 -mmin +120 -name "DIRAC_*" -delete ' % self.workingDirectory
        )
        if status:
            self.log.error("Failure during HTCondorCE __cleanup", stdout)

        # remove all out/err/log files older than "DaysToKeepLogs" days in the working directory
        # not running this for each CE so we do global cleanup
        findPars = dict(workDir=self.workingDirectory, days=self.daysToKeepLogs)
        # remove all out/err/log files older than "DaysToKeepLogs" days
        status, stdout = commands.getstatusoutput(
            r'find %(workDir)s -mtime +%(days)s -type f \( -name "*.out" -o -name "*.err" -o -name "*.log" \) -delete '
            % findPars
        )
        if status:
            self.log.error("Failure during HTCondorCE __cleanup", stdout)
        self._cleanupLock.release()



#!/usr/bin/env python
"""
  Get VM instances available in the configured cloud sites
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

__RCSID__ = "$Id$"

from DIRAC.Core.Base import Script
from DIRAC import gLogger, exit as DIRACExit
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
    Script.setUsageMessage(
        "\n".join(
            [
                "Get VM nodes information",
                "Usage:",
                "%s site ce node [option]... [cfgfile]" % Script.scriptName,
                "Arguments:",
                " cfgfile: DIRAC Cfg with description of the configuration (optional)",
            ]
        )
    )
    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()

    from DIRAC.WorkloadManagementSystem.Client.VMClient import VMClient
    from DIRAC.Core.Security.ProxyInfo import getVOfromProxyGroup

    if len(args) != 3:
        print(Script.showHelp())
        DIRACExit(-1)

    site, ce, node = args

    vmClient = VMClient()
    result = vmClient.stopInstance(site, ce, node)
    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRACExit(-1)

    DIRACExit(0)


if __name__ == "__main__":
    main()



""" DIRAC API Base Class """

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import six
import pprint
import sys

from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.Core.Security.ProxyInfo import getProxyInfo, formatProxyInfoAsString
from DIRAC.Core.Utilities.Version import getCurrentVersion
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getDNForUsername
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSites

__RCSID__ = "$Id$"

COMPONENT_NAME = "API"


def _printFormattedDictList(dictList, fields, uniqueField, orderBy):
    """Will print ordered the supplied field of a list of dictionaries

    :param list dictList: list of dictionaries
    :param list fields: fields
    :param str uniqueField: unique field
    :param str orderBy: ordered
    """
    orderDict = {}
    fieldWidths = {}
    dictFields = {}
    for myDict in dictList:
        for field in fields:
            fieldValue = myDict[field]
            if field not in fieldWidths:
                fieldWidths[field] = len(str(field))
            if len(str(fieldValue)) > fieldWidths[field]:
                fieldWidths[field] = len(str(fieldValue))
        orderValue = myDict[orderBy]
        if orderValue not in orderDict:
            orderDict[orderValue] = []
        orderDict[orderValue].append(myDict[uniqueField])
        dictFields[myDict[uniqueField]] = myDict
    headString = "%s" % fields[0].ljust(fieldWidths[fields[0]] + 5)
    for field in fields[1:]:
        headString = "%s %s" % (headString, field.ljust(fieldWidths[field] + 5))
    print(headString)
    for orderValue in sorted(orderDict):
        uniqueFields = orderDict[orderValue]
        for uniqueField in sorted(uniqueFields):
            myDict = dictFields[uniqueField]
            outStr = "%s" % str(myDict[fields[0]]).ljust(fieldWidths[fields[0]] + 5)
            for field in fields[1:]:
                outStr = "%s %s" % (outStr, str(myDict[field]).ljust(fieldWidths[field] + 5))
            print(outStr)


# TODO: some of these can just be functions, and moved out of here


class API(object):
    """An utilities class for APIs"""

    #############################################################################

    def __init__(self):
        """c'tor"""
        self._printFormattedDictList = _printFormattedDictList
        self.log = gLogger.getSubLogger(COMPONENT_NAME)
        self.section = COMPONENT_NAME
        self.pPrint = pprint.PrettyPrinter()
        # Global error dictionary
        self.errorDict = {}
        self.setup = gConfig.getValue("/DIRAC/Setup", "Unknown")
        self.diracInfo = getCurrentVersion()["Value"]
        self._siteSet = set(getSites().get("Value", []))

    #############################################################################

    def __getstate__(self):
        """Return a copied dictionary containing all the attributes of the API.
            Called when pickling the object. Also used in copy.deepcopy.

        :return: dictionary of attributes
        """
        from DIRAC.FrameworkSystem.private.standardLogging.Logging import Logging

        state = dict(self.__dict__)
        # Replace the Logging instance by its name because it is not copyable
        # because of the thread locks
        if isinstance(state["log"], Logging):
            state["log"] = state["log"].getSubName()
        # In Python 3 PrettyPrinter can't be pickled
        del state["pPrint"]
        return state

    def __setstate__(self, state):
        """Parameter the Job with an attributes dictionary.
            Called when un-pickling the object.

        :params state: attributes dictionary
        """
        self.__dict__.update(state)
        # Build the Logging instance again because it can not be in the dictionary
        # due to the thread locks
        if isinstance(state["log"], six.string_types):
            self.log = gLogger.getSubLogger(state["log"])
        self.pPrint = pprint.PrettyPrinter()

    #############################################################################

    def _errorReport(self, error, message=None):
        """Internal function to return errors and exit with an S_ERROR()

        :param str error: error
        :param str message: message

        :return: S_ERROR(str)
        """
        if not message:
            message = error

        self.log.warn(error)
        return S_ERROR(message)

    #############################################################################

    def _prettyPrint(self, myObject):
        """Helper function to pretty print an object.

        :param object myObject: an object to pring
        """
        print(self.pPrint.pformat(myObject))

    #############################################################################

    def _checkSiteIsValid(self, site):
        """Internal function to check that a site name is valid."""
        if isinstance(site, (list, set, dict)):
            site = set(site) - self._siteSet
            if not site:
                return S_OK()
        elif site in self._siteSet:
            return S_OK()
        return self._reportError("Specified site %s is not in list of defined sites" % str(site))

    #############################################################################

    def _getCurrentUser(self):
        """Get current user

        :return: S_OK(dict)/S_ERROR()
        """
        res = getProxyInfo(False, False)
        if not res["OK"]:
            return self._errorReport("No proxy found in local environment", res["Message"])
        proxyInfo = res["Value"]
        gLogger.debug(formatProxyInfoAsString(proxyInfo))
        if "group" not in proxyInfo:
            return self._errorReport("Proxy information does not contain the group", res["Message"])
        res = getDNForUsername(proxyInfo["username"])
        if not res["OK"]:
            return self._errorReport("Failed to get proxies for user", res["Message"])
        return S_OK(proxyInfo["username"])

    #############################################################################

    def _reportError(self, message, name="", **kwargs):
        """Internal Function. Gets caller method name and arguments, formats the
        information and adds an error to the global error dictionary to be
        returned to the user.

        :param str message: message
        :param str name: name

        :return: S_ERROR(str)
        """
        className = name
        if not name:
            className = __name__
        methodName = sys._getframe(1).f_code.co_name
        arguments = []
        for key in kwargs:
            if kwargs[key]:
                arguments.append("%s = %s ( %s )" % (key, kwargs[key], type(kwargs[key])))
        finalReport = """Problem with %s.%s() call:
Arguments: %s
Message: %s
""" % (
            className,
            methodName,
            "/".join(arguments),
            message,
        )
        if methodName in self.errorDict:
            tmp = self.errorDict[methodName]
            tmp.append(finalReport)
            self.errorDict[methodName] = tmp
        else:
            self.errorDict[methodName] = [finalReport]
        self.log.verbose(finalReport)
        return S_ERROR(finalReport)



#!/bin/env python
"""
Create and put 'ReplicateAndRegister' request.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"
import os
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC import gLogger
import DIRAC


def getLFNList(arg):
    """get list of LFNs"""
    lfnList = []
    if os.path.exists(arg):
        lfnList = [line.split()[0] for line in open(arg).read().splitlines()]
    else:
        lfnList = [arg]
    return list(set(lfnList))


@Script()
def main():
    catalog = None
    Script.registerSwitch("C:", "Catalog=", "Catalog to use")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(" requestName:  a request name")
    Script.registerArgument(" LFNs:         single LFN or file with LFNs")
    Script.registerArgument(["targetSE:     target SE"])
    Script.parseCommandLine()
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "C" or switch[0].lower() == "catalog":
            catalog = switch[1]

    args = Script.getPositionalArgs()

    requestName = None
    targetSEs = None
    if len(args) < 3:
        Script.showHelp(exitCode=1)

    requestName = args[0]
    lfnList = getLFNList(args[1])
    targetSEs = list(set([se for targetSE in args[2:] for se in targetSE.split(",")]))

    gLogger.info(
        "Will create request '%s' with 'ReplicateAndRegister' "
        "operation using %s lfns and %s target SEs" % (requestName, len(lfnList), len(targetSEs))
    )

    from DIRAC.RequestManagementSystem.Client.Request import Request
    from DIRAC.RequestManagementSystem.Client.Operation import Operation
    from DIRAC.RequestManagementSystem.Client.File import File
    from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
    from DIRAC.Core.Utilities.List import breakListIntoChunks

    lfnChunks = breakListIntoChunks(lfnList, 100)
    multiRequests = len(lfnChunks) > 1

    error = 0
    count = 0
    reqClient = ReqClient()
    fc = FileCatalog()
    requestIDs = []
    for lfnChunk in lfnChunks:
        metaDatas = fc.getFileMetadata(lfnChunk)
        if not metaDatas["OK"]:
            gLogger.error("unable to read metadata for lfns: %s" % metaDatas["Message"])
            error = -1
            continue
        metaDatas = metaDatas["Value"]
        for failedLFN, reason in metaDatas["Failed"].items():
            gLogger.error("skipping %s: %s" % (failedLFN, reason))
        lfnChunk = set(metaDatas["Successful"])

        if not lfnChunk:
            gLogger.error("LFN list is empty!!!")
            error = -1
            continue

        if len(lfnChunk) > Operation.MAX_FILES:
            gLogger.error("too many LFNs, max number of files per operation is %s" % Operation.MAX_FILES)
            error = -1
            continue

        count += 1
        request = Request()
        request.RequestName = requestName if not multiRequests else "%s_%d" % (requestName, count)

        replicateAndRegister = Operation()
        replicateAndRegister.Type = "ReplicateAndRegister"
        replicateAndRegister.TargetSE = ",".join(targetSEs)
        if catalog is not None:
            replicateAndRegister.Catalog = catalog

        for lfn in lfnChunk:
            metaDict = metaDatas["Successful"][lfn]
            opFile = File()
            opFile.LFN = lfn
            opFile.Size = metaDict["Size"]

            if "Checksum" in metaDict:
                # # should check checksum type, now assuming Adler32 (metaDict["ChecksumType"] = 'AD'
                opFile.Checksum = metaDict["Checksum"]
                opFile.ChecksumType = "ADLER32"
            replicateAndRegister.addFile(opFile)

        request.addOperation(replicateAndRegister)

        putRequest = reqClient.putRequest(request)
        if not putRequest["OK"]:
            gLogger.error("unable to put request '%s': %s" % (request.RequestName, putRequest["Message"]))
            error = -1
            continue
        requestIDs.append(str(putRequest["Value"]))
        if not multiRequests:
            gLogger.always("Request '%s' has been put to ReqDB for execution." % request.RequestName)

    if multiRequests:
        gLogger.always("%d requests have been put to ReqDB for execution, with name %s_<num>" % (count, requestName))
    if requestIDs:
        gLogger.always("RequestID(s): %s" % " ".join(requestIDs))
    gLogger.always("You can monitor requests' status using command: 'dirac-rms-request <requestName/ID>'")
    DIRAC.exit(error)


if __name__ == "__main__":
    main()



""" RabbitMQSync service to synchronize the internal RabbitMQ database.
    according to CS content. The whole work is done by the RabbitMQSynchronizer
    that is activated when the CS was changed.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC import S_OK
from DIRAC import gConfig
from DIRAC.FrameworkSystem.Utilities import RabbitMQSynchronizer


class RabbitMQSyncHandler(RequestHandler):
    """Service to synchronize the content of internal RabbitMQ database
    with the CS content. The work is done by the RabbitMQSynchronizer
    that acts when the CS is changed.
    """

    @classmethod
    def initializeHandler(cls, _serviceInfo):
        """Handler initialization"""
        syncObject = RabbitMQSynchronizer.RabbitMQSynchronizer()
        gConfig.addListenerToNewVersionEvent(syncObject.sync)
        return S_OK()



#!/usr/bin/env python
########################################################################
# File :    dirac-wms-job-status
# Author :  Stuart Paterson
########################################################################
"""
Retrieve status of the given DIRAC job

Example:
  $ dirac-wms-job-status 2
  JobID=2 Status=Done; MinorStatus=Execution Complete; Site=EELA.UTFSM.cl;
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import os
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch("f:", "File=", "Get status for jobs with IDs from the file")
    Script.registerSwitch("g:", "JobGroup=", "Get status for jobs in the given group")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["JobID:    DIRAC Job ID"], mandatory=False)
    sws, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC import exit as DIRACExit
    from DIRAC.Core.Utilities.Time import toString, date, day
    from DIRAC.Interfaces.API.Dirac import Dirac, parseArguments

    dirac = Dirac()
    exitCode = 0

    jobs = []
    for key, value in sws:
        if key.lower() in ("f", "file"):
            if os.path.exists(value):
                jFile = open(value)
                jobs += jFile.read().split()
                jFile.close()
        elif key.lower() in ("g", "jobgroup"):
            jobDate = toString(date() - 30 * day)
            # Choose jobs no more than 30 days old
            result = dirac.selectJobs(jobGroup=value, date=jobDate)
            if not result["OK"]:
                print("Error:", result["Message"])
                DIRACExit(-1)
            jobs += result["Value"]

    if len(args) < 1 and not jobs:
        Script.showHelp(exitCode=1)

    if len(args) > 0:
        jobs += parseArguments(args)

    result = dirac.getJobStatus(jobs)
    if result["OK"]:
        for job in result["Value"]:
            print("JobID=" + str(job), end=" ")
            for status in result["Value"][job].items():
                print("%s=%s;" % status, end=" ")
            print()
    else:
        exitCode = 2
        print("ERROR: %s" % result["Message"])

    DIRACExit(exitCode)


if __name__ == "__main__":
    main()



""" Base class for plugins as used in the transformation system
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re

from DIRAC import S_ERROR, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations


class PluginBase(object):
    """Base class for TS plugins"""

    def __init__(self, plugin, operationsHelper=None):
        """plugin name has to be passed in: it will then be executed as one of the functions below, e.g.
        plugin = 'BySize' will execute TransformationPlugin('BySize')._BySize()
        """
        self.plugin = plugin
        self.params = {}

        if operationsHelper is None:
            self.opsH = Operations()
        else:
            self.opsH = operationsHelper

    def setParameters(self, params):
        """Extensions may re-define it"""
        self.params = params

    def run(self):
        """this is a wrapper to invoke the plugin (self._%s()" % self.plugin)"""
        try:
            evalString = "self._%s()" % self.plugin
            return eval(evalString)  # pylint: disable=eval-used
        except AttributeError as x:
            if re.search(self.plugin, str(x)):
                return S_ERROR("Plugin not found")
            else:
                gLogger.exception("Exception in plugin", self.plugin, lException=x)
                raise AttributeError(x)
        except Exception as x:
            gLogger.exception("Exception in plugin", self.plugin, lException=x)
            raise Exception(x)
